/*
Copyright (c) 1997-2000 Allaire.  All Rights Reserved.

DO NOT REDISTRIBUTE THIS SOFTWARE IN ANY WAY WITHOUT THE EXPRESSED
WRITTEN PERMISSION OF ALLAIRE.
*/

/* Include mod_jrun header */
#include "mod_jrun.h"
#include "jrun_proxy.h"
#include "jrun_property.h"

#ifndef JRUN_MODULE
#define JRUN_MODULE jrun_module
#endif

/* Forward declaration for module object. */
#ifdef WIN32
MODULE_VAR_EXPORT module JRUN_MODULE;
#else
module JRUN_MODULE;
#endif

/* verbose mode */
static int Verbose = FALSE;
 
/* routines in this file */
static void jrun_exit(int sig);

/* Init JRun module */
static void 
jrun_init(server_rec *s, pool *p)
{
	jrun_request r;
	init_apache_request(&r, NULL, s);
	r.log(&r, "%s %s Apache module - %s %s", 
		JRUN_APP_NAME, JRUN_VERSION, __DATE__, __TIME__);
}

/* Init JRun child process */
static void 
jrun_child_init(server_rec *s, pool *p)
{
    jrun_conf * c;
	jrun_request r;

	/*   Create connections to JRun Connector Proxy servers */
	do {
		init_apache_request(&r, NULL, s);

		c = get_module_config( s->module_config, &JRUN_MODULE );
		initProxy(&c->pxy, &r, c->jrunroot, c->jvmlist);
		if (c->jrun_host != null && c->rulesPath != null) {
			char jrunproxy[512];
			sprintf(jrunproxy, "%.20s:%d;%s", c->jrun_host, c->jrun_port, c->rulesPath);
			c->pxy.addServer(&c->pxy, &r, jrunproxy);
		}
		s = s->next;
	} while(s != NULL);
}

/* Exit JRun child process */
static void 
jrun_child_exit(server_rec *s, pool *p)
{
    jrun_conf * c;
	jrun_request r;

	/* Close connections to JRun Connector Proxy server */
	do {
		init_apache_request(&r, NULL, s);
		r.trace(&r, "jrun_child_exit");
		c = get_module_config( s->module_config, &JRUN_MODULE );
		c->pxy.destroy(&c->pxy, &r);
		s = s->next;
	} while(s != NULL);
}


/* Make a config record. */
static void*
make_jrun_conf( pool *p, server_rec *s )
{
	jrun_request r;
    jrun_conf* c = (jrun_conf*) pcalloc( p, sizeof(jrun_conf) );

	init_apache_request(&r, NULL, s);
	c->jrun_host	= null;
    c->jrun_port 	= JRUN_PROXY_PORT;
    c->rulesPath 	= null;
	c->jrunroot = null;
	c->jvmlist = null;
#if MODULE_MAGIC_NUMBER <= 19980401
	/* special flag for initializing Apache 2.X child processes */
	c->first_time = 1;
#endif

	r.trace(&r, "make_jrun_conf made");

    return (void*) c;
}

/* JRunConfig command. */
static const char *
jrun_config( cmd_parms* parms, void* dummy, char* var, char* val )
{
	jrun_request jr;
    server_rec* s = parms->server;
    jrun_conf* c = get_module_config( s->module_config , &JRUN_MODULE );

	init_apache_request(&jr, NULL, s);
    if ( strcasecmp( var, "Verbose" ) == 0 )
	{
		if(strcasecmp(val, "true") == 0)
			Verbose = TRUE;
		else
			Verbose = FALSE;
    }
    else if ( strcasecmp( var, "Mappings" ) == 0 )
	{
		jr.trace(&jr, "JRunConfig: mappings property file: %s", val);
		c->rulesPath = val;
	}
	else if ( strcasecmp( var, "ProxyHost" ) == 0 || strcasecmp( var, "JRunProxyHost" ) == 0 )
		c->jrun_host = val;
    else if ( strcasecmp( var, "ProxyPort" ) == 0 || strcasecmp( var, "JRunProxyPort" ) == 0)
		c->jrun_port = atoi(val);
	else if (strcasecmp(var, "errorurl") == 0)
		c->errorUrl = val;
	else if (strcasecmp(var, "jrun.rootdir") == 0) {
		c->jrunroot = val;
	} else if (strcasecmp(var, "jvmlist") == 0) {
		c->jvmlist = val;
	}
    return (char*) 0;
}

static int
jrun_trans(request_rec * r)
{
	jrun_conf* c;
	server_rec* s = r->server;
	jrun_request jr;
	jrun_proxy *target;

	/* Init request object */
	init_apache_request(&jr, r, s);

	c = get_module_config( r->server->module_config, &JRUN_MODULE );

	/* apache 1.2.X didn't call the child init function */
#if MODULE_MAGIC_NUMBER <= 19980401
	if (c->first_time == 1) {
		c->first_time = 0;
		jrun_child_init(r->server, r->pool);
	}
#endif

	jr.path = r->filename;
	target = c->pxy.filterRequest(&c->pxy, &jr);
	if (target != NULL) {
		r->handler = pstrdup( r->pool, "jrun-handler" );
		/* save the proxy pointer for later when we're in jrun_service.
		   use table_setn so apache doesn't try to strdup the target pointer */
		table_setn(r->notes, "jrun-proxy-target", (const char *) target);
		return OK;
	}

	return DECLINED;
}


/* return an array of the HTTP request headers */
static char **
getHeaderNames(jrun_request *jr)
{
	char *empty[1] = { NULL }, **names;
	request_rec *r = (request_rec *) jr->request;
	int i, j, count;
	table_entry *hdrNames;

	count = table_elts(r->headers_in)->nelts;
	hdrNames = (table_entry *) table_elts(r->headers_in)->elts;
	names = jr->malloc(jr, (1+count) * sizeof(char *));

	for (i=j=0; i < count; i++) {
		if (hdrNames[i].key != NULL) {	
			names[j++] = jr->jstrdup(jr, hdrNames[i].key);
		}
	}
	while (j <= count) {
		names[j++] = NULL;
	}
	return names;
}

/* free an array of strings */
static void freeList(jrun_request *jr, char **v)
{
	int i;
	if (v != NULL) {
		for (i=0; v[i] != NULL; i++) {
			jr->free(jr, v[i]);
		}
		jr->free(jr, v);
	}
}


/* ED 3/11/98 need conditionals for old apache's */
#if MODULE_MAGIC_NUMBER <= 19980401 && !defined(APACHE_SSL)
#define http_method(req) "http"
#endif

/* Request handler.  Called on each request. */
static int
jrun_service( request_rec* req )
{
	jrun_request jr;
	server_rec* s = req->server;
	jrun_conf* c = get_module_config( s->module_config, &JRUN_MODULE );
	int error;
	jrun_proxy *target;
	char *h;

	/* Init request object */
	init_apache_request(&jr, req, s);

	jr.trace(&jr, "enter jrun_service");

	/* add additional headers */
	add_cgi_vars(req);
	add_common_vars(req);

	/* apache doesn't put Authorization into subprocess_env for
	security.  we needed this because JRun often asks for AUTHORIZATION
	and we didn't have the case-insensitive search.  This should not
	be needed as long as we search headers_in first and don't match case */

    h = (char *) table_get(req->headers_in, "Authorization" ) ;
	if (h != NULL)
		table_set(req->subprocess_env, "AUTHORIZATION", h);

	/* ED 3/4/99 check whether we are running secure */
	if (strcasecmp(http_method(req),"https") == 0) {
		table_set(req->subprocess_env, "SERVER_PORT_SECURE", "1");
	} else {
		table_set(req->subprocess_env, "SERVER_PORT_SECURE", "0");
	}

	target = (jrun_proxy *) table_get(req->notes, "jrun-proxy-target");
	if (target == null) {
		target = &c->pxy;
	} else {
		jr.trace(&jr, "jrun_service found jrun-proxy-target %s", target->name);
	}

	/* set up for reading post data, if any */
	if ((error = setup_client_block(req, REQUEST_CHUNKED_ERROR)) != OK) {
		jr.log(&jr, "setup_client_block failed %d", error);
		return error;
	}
	
	jr.shouldRead = 1;
	jr.shouldReadMore = 1;

	jr.hdrNames = getHeaderNames(&jr);

	/* handle the request with the proxy we found */
	if (target->handleRequest(target, &jr, NULL ) == -1 ) {
		freeList(&jr, jr.hdrNames);
		jr.trace(&jr, "handleRequest failed");
		return 500;
	} else {
		freeList(&jr, jr.hdrNames);
		jr.trace(&jr, "jrun_service returning %d", req->status);
		return OK; /* return req->status; */
	}
}

/* Command table. */
static command_rec jrun_commands [] =
{
    { 
		"JRunConfig" , 
		jrun_config, 
		NULL, 
		ACCESS_CONF|RSRC_CONF, TAKE2, 
		"a variable to set followed by its value" },
    NULL
};

/* Handler record. */
static handler_rec jrun_handlers[] =
{
    { "jrun-handler", jrun_service }, 
	{ NULL }
};

/* Module structure. */
#ifdef WIN32
module MODULE_VAR_EXPORT JRUN_MODULE =
#else
module JRUN_MODULE =
#endif
{
    STANDARD_MODULE_STUFF,
    jrun_init,				/* initializer */
    NULL,					/* dir config creater */
    NULL,					/* dir merger - default is to override */
    make_jrun_conf,			/* server config */
    NULL,					/* merge server config */
    jrun_commands,			/* command table */
    jrun_handlers,			/* handlers */
    NULL,				/* filename translation */
    NULL,					/* check_user_id */
    NULL,					/* check auth */
    NULL,					/* check access */
    jrun_trans,					/* type_checker */
    NULL,					/* fixups */
    NULL,					/* [10] logger */
    NULL,					/* [3] header parser */
#if MODULE_MAGIC_NUMBER > 19980401
    jrun_child_init,		/* process initializer */
    jrun_child_exit,		/* process exit/cleanup */
    NULL					/* [1] post read_request handling */
#endif
};


/* the old interface.c - 2/17/99 - tpr */

static char *
getRealPath(jrun_request *jr, const char *path)
{
	request_rec *r = (request_rec *) jr->request;
	request_rec *pa_req;
	char *esc_path, *value;
	
	esc_path = escape_uri(r->pool, path);
	jr->trace(jr, "   ap_escape_uri(%s) = %s", path, esc_path);
	pa_req = sub_req_lookup_uri(esc_path, r);

	value = pstrcat(r->pool, pa_req->filename, pa_req->path_info, NULL);
	destroy_sub_req(pa_req);
	jr->trace(jr, "   getRealPath(%s) = %s", path, value?value:"<null>");

	return (value == NULL ? NULL : jr->jstrdup(jr, value));
}

static char *
getHeader(jrun_request *jr, const char *headerName)
{
	request_rec *r = (request_rec *) jr->request;
	char *buf = NULL, *value;

	if(strcasecmp(headerName, "REMOTE_HOST") == 0)
		headerName = "REMOTE_ADDR";

	value = (char *) table_get(r->headers_in, headerName);
	if (value == NULL) {
		value = (char *) table_get(r->subprocess_env, headerName);
		if (value == NULL) {
			buf = jr->malloc(jr, strlen(headerName) + 6);
			if (buf != NULL) {
				strcpy(buf, "HTTP_");
				strcat(buf, headerName);
				value = (char *) table_get(r->subprocess_env, buf);
				if (value != NULL) {
					headerName = buf;
				}
			}
		}
	}

	jr->trace(jr, "   %s: %s", headerName, value ? value : "<null>");
	if (buf != NULL) {
		jr->free(jr, buf);
	}
	return (value == NULL ? NULL : jr->jstrdup(jr, value));
}

static int
getPostData(jrun_request *jr, char *buf, int clen)
{
	request_rec *r = (request_rec *) jr->request;
	int len = 0;

	/* setup_client_block has already been called, and shouldRead flags
		are pre-set based on method */
	if (jr->shouldRead && should_client_block(r) || jr->shouldReadMore) {
		jr->shouldRead = 0;

		/* 0=eof, -1=error, >0=success */
		len = get_client_block(r, buf, clen);
		if (len <= 0) {
			jr->shouldReadMore = 0;
		} else {
			jr->trace(jr, "   read %d bytes of post.", len);
			jr->shouldReadMore = 1;
		}
	}
	jr->trace(jr, "   getPostData returning %d", len);
	return len;
}

/* create a lower-case copy of a header value.  mainly used for content-type,
   content-encoding, etc that have to be lowercase.
   ED 5/3/99 stop at first ';' to correctly handle this case:
             "content-type: multipart/x-mixed-replace; <boundary string>"
*/

static char *pstrdup_lower(request_rec *r, const char *s)
{
	char *p, *new = pstrdup(r->pool, s);
	if (new != NULL) {
		for (p = new; *p != '\0' && *p != ';'; p++)
			*p = tolower(*p);
	}
	return new;
}


static void
setStatus(jrun_request *jr, int statusCode, const char *szStatus)
{
	request_rec *r = (request_rec *) jr->request;
	char *status_line;

	/* set the status line */
	r->status = statusCode;
	/* ED 3/23/99 r->status_line is (const char *) in Apache 136 */
	status_line = palloc(r->pool, strlen(szStatus) + 20);
	sprintf(status_line, "%d %s", statusCode, szStatus);
	r->status_line = status_line;

	if (statusCode == 503) {
		server_rec *s = r->server;
	    jrun_conf* c = get_module_config( s->module_config , &JRUN_MODULE );
		if (c != NULL && c->errorUrl != NULL) {
			r->status = 302;
			sprintf(status_line, "302 %s", szStatus);
			set_content_length(r, 0);
			table_set(r->headers_out, "Location", c->errorUrl);
			jr->noaction = 1;
			jr->trace(jr, "   redirect ap_send_http_header()");
			send_http_header(r);
			return;
		}
	}
}

static void
addHeader(jrun_request *jr, const char *name, const char *value)
{
	request_rec *r = (request_rec *) jr->request;
		if (strcasecmp(name, "content-length") == 0) {
			jr->trace(jr, "   ap_set_content_length(%d)", atoi(value));
			set_content_length(r, atoi(value));

		/* do nothing -- apache does keepalive automatically
		} else if (strcasecmp(name, "connection") == 0 &&
						strcasecmp(value, "keep-alive") == 0) {
			jr->trace(jr, "   ap_set_keepalive()"); */

		} else if (strcasecmp(name, "content-type") == 0) {
			jr->trace(jr, "   request_rec->content_type = %s", value);
			r->content_type = pstrdup_lower(r, value);

		} else if (strcasecmp(name, "content-encoding") == 0) {
			jr->trace(jr, "   request_rec->content_encoding = %s", value);
			r->content_encoding = pstrdup_lower(r, value);

		/* ED 3/3/99 cookies can be repeated */
		} else if (strcasecmp(name, "set-cookie") == 0) {
			jr->trace(jr, "   adding cookie %.50s: %.200s", name, value);
			table_add(r->headers_out, name, value);

		} else {
			/* put it in r->headers_out */
			table_set(r->headers_out, name, value);
			jr->trace(jr, "   table_set(%s, %s)", name, value);
		}
}


static int
writeClient(jrun_request *jr, const char *data, int len)
{
	request_rec *r = (request_rec *) jr->request;

	if (!jr->noaction && !jr->committed) {
		jr->trace(jr, "   ap_send_http_header()");
		send_http_header(r);
	}

	if (!jr->noaction) {
		if (len > 0) {
			jr->trace(jr, "   ap_rwrite(%d bytes)", len);
			if( rwrite(data, len, r) == EOF )
				return -1;
		}
		rflush(r);
	}

	return 0;
}

static void *apache_malloc(jrun_request *r, int len)
{
	request_rec *req = (request_rec *) r->request;
	void *result = palloc(req->pool, len <= 0 ? 1 : len);
	if (result == null) {
		r->log(r, "apache_malloc(%d) failed!", len);
	}
	return result;
}

static void apache_free(jrun_request *r, void *p)
{
	/* don't need to free pooled memory */
}


/* init jrun_request object with function pointers */
static void
init_apache_request(jrun_request *r,  request_rec *req, server_rec *serv)
{
	init_jrun_request(r);
	r->request = (void *) req;
	r->context = (void *) serv;

	if (req == NULL || req->connection == NULL) {
		sprintf(r->stringRep, "jrApache[init] ");
	} else {
		sprintf(r->stringRep, "jrApache[%d] ",
			ntohs(req->connection->remote_addr.sin_port));
	}

	r->getHeader = getHeader;
	r->getRealPath = getRealPath;
	r->getPostData = getPostData;
	r->setStatus = setStatus;
	r->addHeader = addHeader;
	r->writeClient = writeClient;

	r->log = WriteToLog;
	if (Verbose)
		r->trace = WriteToLog;

	if (req != NULL) {
		r->uri = req->uri;
		r->malloc = apache_malloc;
		r->free = apache_free;
	}
}


static void
WriteToLog(jrun_request *r, const char *szFormat, ...)
{
	server_rec *s = (server_rec *) r->context;
    va_list list;
    char szBuf[2048];

	strcpy(szBuf, r->stringRep);
    va_start (list, szFormat);
    vsprintf (strchr(szBuf,'\0'), szFormat, list);
    va_end (list);

#if MODULE_MAGIC_NUMBER > 19980401
	/* don't need to add newline - this function does it for us */
    ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_NOTICE, s, "%s", szBuf);
#else
    log_error(szBuf, s);
#endif

#ifdef WIN32
	strcat(szBuf, "\r\n");
	OutputDebugString(szBuf);
#endif
}
