/*
Copyright (c) 1997-2000 Allaire.  All Rights Reserved.

DO NOT REDISTRIBUTE THIS SOFTWARE IN ANY WAY WITHOUT THE EXPRESSED
WRITTEN PERMISSION OF ALLAIRE.
*/

#include <stdarg.h>
#include "jrun_property.h"

static char *expand_vars(const char *s, const propTable *props, const propTable *defaultProps);
static char *expand_vars2(const char *s, const propTable *props, const propTable *defaultProps,
						  const propTable *jvmProps, const char *jvm);

/* malloc/free functions for use by loadProperties1 only */
static void *
doMalloc(jrun_request *jr, int len)
{
	void *result = malloc(len <= 0 ? 1 : len);
	return result;
}

static char *
doStrdup(jrun_request *jr, const char *s)
{
	char *result = strdup(s?s:"");
	return result;
}

static void
doFree(jrun_request *jr, void *p)
{
	if(p) free(p);
}

static void
doTrace(jrun_request *r, const char *szFormat, ... )
{
#ifdef WIN32
	char szBuf[1024];
	va_list list;

	strcpy(szBuf, r->stringRep);
	va_start (list, szFormat);
	vsprintf (strchr(szBuf, '\0'), szFormat, list);
	va_end (list);
	strcat(szBuf, "\r\n");

	// Send to dbmon
	OutputDebugString(szBuf);
#endif
}

static void init_allocators(jrun_request *r)
{
	memset( r, 0, sizeof( *r ) );
	r->malloc = doMalloc;
	r->free = doFree;
	r->jstrdup = doStrdup;
	r->perm_malloc = doMalloc;
	r->perm_free = doFree;
	r->trace = doTrace;
	r->log = doTrace;
}

propTable *
loadProperties1(const char *filename)
{
	jrun_request r;

	init_allocators(&r);
	return loadProperties( &r, filename );
}

/* parse a full property file, including comments, etc.  String is assumed
to be null terminated */
propTable *parseProperties(jrun_request *r, const char *propString)
{
	int currProp = 0;
	int propCount = 10;
	const signed char *p, *s;
	propTable * props = (propTable *) r->perm_malloc(r, sizeof(propTable));
	int c;
	char *name, *value, *v;

	if (props == null)
		return null;

	/* set the count to zero */
	props->count = 0;

	p = (const signed char *) propString;
	for (c = (int) *p++; c != '\0'; c = (int) *p++) {
		if(c == 0 || c == 26) {  /* EOF or ^Z */
			break;
		} else if(c == '#') {
			/* comment */
			while ((c = (int)*p++) != 0 && c != '\n')
				;
		} else if(c == '\r' || c == '\n') {
			/* blank line */
		} else {
			s = p-1; /* name starting point */
			while( (c = (int)*p++) != '=' && c != '\r' && c != '\n') {
				if (c == 0)		return props;
			}
			if (c == '\r' || c == '\n') {
				r->trace(r, "property %.*s has no value. skipping.",
					p-s-1, s);
				continue;
			}
			/* p points to first character after the '=' */
			name = r->perm_malloc(r, p-s);
			strncpy(name, (const char *) s, p-s-1);
			name[p-s-1] = '\0';

			s = p;
			while ((c = (int)*p++) != '\r' && c != '\n' && c != '\0') {
			}
			while (c == '\r') c = (int)*p++; /* get to EOL */
			v = value = r->perm_malloc(r, p-s);
			/* copy string and fix escapes along the way */
			while (s < p-1) {
				if (*s == '\\') s++;
				if (s < p-1) *v++ = *s++;
			}
			*v = '\0';
			r->trace(r, "  %.100s = %.200s", name, value);
			addProperty(props, name, value);
			if (c == '\0') {
				break; /* no EOL at end of file */
			}
		}
	}
	return props;
}

/*
 * nonzero pointer to props table = success
 * null = failure
 */
propTable *
loadProperties(jrun_request *r, const char *filename)
{
	FILE *fp;
	propTable * props = null;
	size_t length;
	extern int errno;

	/* read the property file */
	r->trace(r, "loadProperties(%s)", filename);
	if((fp = fopen(filename, "r")) != null) {
		if (fseek(fp, 0, SEEK_END) == 0)
			length = ftell(fp);
		if (length > 0) {
			char *data = r->malloc(r, length + 1);
			fseek(fp, 0, SEEK_SET);
			if ((length = fread(data, 1, length, fp)) > 0) {
				data[length] = '\0';
				props = parseProperties(r, data);
				r->free(r, data);
			} else {
				r->log(r, "could not read data in %s", filename, length);
			}
		}
		fclose(fp);
	} else {
		r->log(r, "could not open %s: %s", filename, strerror(errno));
	}
	return props;
}

void freeProperties(jrun_request *r, propTable *props)
{
	int i;

	if (props != null) {
		for (i=0; i < props->count; i++) {
			if (props->entry[i].name != null)
				r->perm_free(r, props->entry[i].name);
			if (props->entry[i].value != null)
				r->perm_free(r, props->entry[i].value);
		}
		r->perm_free(r, props);
	}
}

const char *
getPropertyValue(const propTable *props, const char *name)
{
	int i;
	char *propName;

	if( props == null )
		return null;

	for(i=0;i<props->count;i++){
		propName = props->entry[i].name;
		if(strcmp(name, propName) == 0)
			return props->entry[i].value;
	}

	return null;
}

const char *
getPropertyValue2(const propTable *props, const propTable *defaultProps, const char *name)
{
	const char *ret = getPropertyValue(props, name);

	if((ret == NULL || !strcmp(ret, "{default}")) && defaultProps != NULL)
		ret = getPropertyValue(defaultProps, name );
	if(ret != NULL)
		ret = expand_vars((const char *)ret, props, defaultProps); 
	return ret;
}

const char *
getPropertyValue3(const propTable *props, const propTable *defaultProps, const propTable *jvmProps,
				  const char *name, const char *jvm)
{
	const char *ret = getPropertyValue(props, name);

	if((ret == NULL || !strcmp(ret, "{default}")) && defaultProps != NULL)
		ret = getPropertyValue(defaultProps, name );
	if(ret != NULL)
		ret = expand_vars2((const char *)ret, props, defaultProps, jvmProps, jvm); 
	return ret;
}

static int 
find_variable(const char *s, char **start, char **end)
{
	*start = strchr(s, '{');
	if(*start != NULL)
		*end = strchr(*start, '}');
	return *start != NULL && *end != NULL;
}

/*
 * expand_vars -- expand property variables enclosed within {}'s.
 */
static char *
expand_vars(const char *s, const propTable *props, const propTable *defaultProps)
{
	char *ts, *te; 
	const char *expansion;
	char *buf;
	char *buf2;

	buf = strdup(s);
	while(find_variable(buf, &ts, &te)) {
		*ts++ = 0;
		*te = 0;
		expansion = getPropertyValue2(props, defaultProps, ts);
		if(expansion == NULL){
			buf2 = malloc(strlen(buf) + strlen(te+1) + 1);	
			sprintf(buf2, "%s%s", buf, te+1);
		} else {
			buf2 = malloc(strlen(buf) + strlen(expansion) + strlen(te+1) + 1);	
			sprintf(buf2, "%s%s%s", buf, expansion, te+1);
		}
		free(buf);
		buf = buf2;
	}
	return buf;

}


/*
 * expand_vars2 -- expand property variables enclosed within {}'s.
 */
static char *
expand_vars2(const char *s, const propTable *props, const propTable *defaultProps,
			 const propTable *jvmProps, const char *jvm)
{
	char *ts, *te; 
	const char *expansion;
	char *buf;
	char *buf2;

	buf = strdup(s);
	while(find_variable(buf, &ts, &te)){
		*ts++ = 0;
		*te = 0;
		if (strcmp(ts, "jrun.server.rootdir") == 0) {
			expansion = getPropertyValue(jvmProps, jvm);
		} else if (strcmp(ts, "jrun.server.name") == 0) {
			expansion = strdup(jvm);
		} else {
			expansion = getPropertyValue3(props, defaultProps, jvmProps, ts, jvm);
		}
		if(expansion == NULL){
			buf2 = malloc(strlen(buf) + strlen(te+1) + 1);	
			sprintf(buf2, "%s%s", buf, te+1);
		} else {
			buf2 = malloc(strlen(buf) + strlen(expansion) + strlen(te+1) + 1);	
			sprintf(buf2, "%s%s%s", buf, expansion, te+1);
		}
		free(buf);
		buf = buf2;
	}
	return buf;

}

void
setProperty(jrun_request *r, propTable *props,
				 const char *name, const char *value)
{
	int i;
	char *propName;
	jrun_request simpleRequest; /* just for the allocators */
	if (r == null) {
		r = &simpleRequest;	
		init_allocators(r);
	}
	for(i=0;i<props->count;i++) {
		propName = props->entry[i].name;
		if (strcmp(name, propName) == 0) {
			r->free(r, props->entry[i].value);
			props->entry[i].value = r->perm_malloc(r, strlen(value)+1);
			strcpy(props->entry[i].value, value);
			return;
		}
	}
	/* TODO: make sure we don't overflow */

	props->entry[i].name = r->perm_malloc(r, strlen(name)+1);
	strcpy(props->entry[i].name, name);
	props->entry[i].value = r->perm_malloc(r, strlen(value)+1);
	strcpy(props->entry[i].value, value);
	props->count++;
}

void addProperty(propTable *props, char *name, char *value)
{
	props->entry[props->count].name = name;
	props->entry[props->count].value = value;
	props->count++;
}

/* warning -- this is not threadsafe, but not dangerous either */
char *describeError(int err)
{
	static char message[256];
	char *s;
	switch (err) {
#ifdef _WINSOCKAPI_
	case WSAEINTR: return "Interrupted system call";
	case WSAEACCES: return "Permission denied";
	case WSAEWOULDBLOCK: return "Operation would block";
	case WSAEINPROGRESS: return "Operation now in progress";
	case WSAEALREADY: return "Operation already in progress";
	case WSAEADDRINUSE: return "Address already in use";
	case WSAEADDRNOTAVAIL: return "Cannot assign requested address";
	case WSAENETDOWN: return "Network is down";
	case WSAENETUNREACH: return "Network is unreachable";
	case WSAENETRESET: return "Network dropped connection because of reset";
	case WSAECONNABORTED: return "Connection aborted";
	case WSAECONNRESET: return "Connection reset by peer";
	case WSAENOBUFS: return "No buffer space available";
	case WSAEISCONN: return "Socket is already connected";
	case WSAENOTCONN: return "Socket is not connected";
	case WSAESHUTDOWN: return "Can't send after socket shutdown";
	case WSAETIMEDOUT: return "Connection timed out";
	case WSAECONNREFUSED: return "Connection refused";
	case WSAEHOSTDOWN: return "Host is down";
	case WSAEHOSTUNREACH: return "No route to host";
#endif
	case EJRUN_BUSY: return "JRun too busy or out of memory";
	case EJRUN_PROTOCOL_ERROR: return "JRun connector protocol out of sync";
	default:
		s = strerror(err);
		sprintf(message, "%d %s", err, (s != null) ? s : "null");
	}
	return message;
}
