/*
Copyright (c) 1997-2000 Allaire.  All Rights Reserved.

DO NOT REDISTRIBUTE THIS SOFTWARE IN ANY WAY WITHOUT THE EXPRESSED
WRITTEN PERMISSION OF ALLAIRE.
*/

#include "jrun_property.h"
#include "jrun_proxy_headers.h"

#include <time.h>
#include <ctype.h>

/* MUTEX code.  Maybe this should go into the API specific code, but since
most of it would just be replicated, we will save ourselves some work and
leave it here, with #ifdef's for different platforms */

static void
initSync(jrun_request *r, jrun_proxy *pxy)
{
#if defined(XP_WIN32) || defined(XP_UNIX) || defined(XP_NETWARE)
	pxy->lock = crit_init();
#elif defined(WIN32)
	InitializeCriticalSection(&pxy->lock);
#endif /* !WIN32 */
	pxy->locked = 0;
}

static void
startSync(jrun_request *r, jrun_proxy *pxy)
{
#if defined(XP_WIN32) || defined(XP_UNIX) || defined(XP_NETWARE)
	if (r->locked == 1)
		r->log(r, "caught nested lock");
	crit_enter(pxy->lock);
#elif defined(WIN32)
	EnterCriticalSection(&pxy->lock);	
#endif /* !WIN32 */
	if (pxy->locked == 1)
		r->log(r, "startSync - detected broken mutex!");
	pxy->locked = 1;
	r->locked = 1;
}

static void
stopSync(jrun_request *r, jrun_proxy *pxy)
{
	if (pxy->locked == 0)
		r->log(r, "stopSync - detected broken mutex!");
	pxy->locked = 0;
	r->locked = 0;
#if defined(XP_WIN32) || defined(XP_UNIX) || defined(XP_NETWARE)
	crit_exit(pxy->lock);
#elif defined(WIN32)
	LeaveCriticalSection(&pxy->lock);	
#endif /* !WIN32 */
}

/**
 * sendBytes 
 * 
 * sends "avail" to socket
 *
 * return <number of bytes> == success, -1 == failure
 */
static int
sendBytes(jrun_request *r, const void *bytes, int avail, int sock)
{
	int result = 1;
	int sent = 0;

	while ((sent < avail) && (result > 0)) {
		result = r->send(r, sock, (const char *) bytes + sent, avail - sent);
		if (result > 0) 
			sent += result;
		if (sent < result)
			r->trace(r, "sent %d/%d",sent,avail);
	}

	if(sent != avail)
		return -1;
	else
		return sent;
}

/**
 * readFully 
 * 
 * reads in len number of bytes from socket
 *
 * @return 0 == success, -1 == failure
 */
static int
readFully(jrun_request *r, void *bytes, int len, int sock)
{
	int read = 1;
	int total = 0;

	while ((total < len) && (read > 0)) {
		read = r->recv(r, sock, (char *) bytes + total, len - total);
	    if (read > 0) 
			total += read;
	}

	if(total != len)
		return -1;
	else
		return 0;
}

static void free_jrun_request(jrun_request *r)
{
	if (r->reason != null) r->free(r, r->reason);
	if (r->headers != null) r->free(r, r->headers);
	r->reason = r->headers = r->headersEnd = NULL;
}

/*
 * initPool - initialize the socket pool.  The pool is a stack of socket
 * handles.  The global mutex locks are used to ensure safe access to the
 * data structure.  All sockets in the stack are connected to the host and
 * port given.  Host must be in numeric format.
 */
static sockPool *
initPool(jrun_request *r, jrun_proxy *pxy,
		 const char *host, int port)
{
	sockPool *spool = (sockPool *) r->perm_malloc(r, sizeof(sockPool));

	r->trace(r, "initPool");

	/* future -- do hostname lookup here */
	memset(&spool->addr, 0, sizeof(spool->addr));  
	spool->addr.sin_family = AF_INET;
	spool->addr.sin_addr.s_addr = inet_addr(host);
	spool->addr.sin_port = htons((unsigned short) port);

	/* pool semantics:
	   connections that are in the linked list (stack) are available
	   for re-use.  Busy connections are removed from the stack before
	   being used. in either case, sock may or may not be valid.
	*/

	spool->pool = null;
	r->trace(r, "%s initPool finished", pxy->name);

	return spool;
}

static void
returnJRunConnectionToPool(jrun_request *r, jrun_proxy *pxy, poolEntry * p)
{
	sockPool *spool = (sockPool *) pxy->obj1;

	free_jrun_request(r);
	startSync(r, pxy);
	p->next = spool->pool;
	spool->pool = p;
	stopSync(r, pxy);
	r->trace(r, "%s returned [%d] to pool", pxy->name, p->sock);
}

static poolEntry *
getJRunConnectionFromPool(jrun_proxy *pxy, jrun_request *r)
{
	sockPool *spool = (sockPool *) pxy->obj1;
	poolEntry *e = null;

	/* see if we have any idle connections already saved up */
	startSync(r, pxy);
	if (spool->pool != null) {
		e = spool->pool;
		spool->pool = e->next;
	}
	stopSync(r, pxy);

	if (e != null) {
		/* this socket was already open, so be sensitive to problems */
		e->next = e; /* any value but null means the socket was pooled */
	} else {
		/* need a new connection */
		int sock = r->connect(r, &spool->addr);
		if (sock != -1) {
			e = r->perm_malloc(r, sizeof(poolEntry));
			e->next = null;
			e->sock = sock;
		}
	}
	return e;
}

/*
 * destroyJRunConnection - close a socket and free its data structure. 
 * Don't need to mess around with the stack, because we're not saving
 * this connection.
 */
static void
destroyJRunConnection(jrun_request *r, jrun_proxy *pxy, poolEntry *p)
{
	/* we are destroying a connection to jrun, so mark the mappings
	   stale so we will refresh them next chance we get */
	pxy->next_mapcheck_time = time(0)-1;
	r->close(r, p->sock );
	r->perm_free(r, p);
	free_jrun_request(r);
}

/* 
 * readData
 *
 * returns pointer to string and sets strlen.
 * if any error occurs, sets r->errno and returns null.
 */
static char *
readData(int sock, jrun_request *r, int *strlen)
{
	int len;
	char *val;

	if( readFully(r, &len, sizeof(int), sock) == -1 ) {
		/* socket error */
		return null;
	}

	len = ntohl(len);
	val = (char *) r->malloc(r, (len + 1));

	if(val == null)	{
		/* memory error */
		r->error = ENOBUFS;
	} else 	if( readFully(r, val, len, sock) == -1) {
		/* socket error */
		r->free(r, val);
		val = null;
	} else {
		/* success */
		val[len] = '\0';
		if (strlen != null)
			*strlen = len;
	}

	return val;
}

/* 
 * writeString
 *
 * return 0 == success, -1 == failure
 *
 */
static int
writeString(jrun_request *r, int sock, char *value)
{
	int slen = 0;
	int *buf;
	
	/* protect ourself from a  null string by sending a 0-length chunk */
	if (value == null) value = "";
	slen = strlen(value);
	
	/* allocate a block of memory so we can send the whole command in one chunk */
	buf = r->malloc(r, slen + sizeof(int));
	if (buf == null)
		return -1;
	buf[0] = htonl(slen);   /* command length */
	if (slen > 0) {
		/* command data */
		memcpy(&buf[1], value, slen);
	}
	
	if( sendBytes(r, buf, slen + sizeof(int), sock) == -1 ) {
		r->free(r, buf);
		return -1;
	} else {
		r->free(r, buf);
		return 0;
	}
}


/* 
 * sendRequest
 *
 * return -1 = failure, anything else = success
 *
 */

#define appendInt(p, off, val) \
		(((p)[off++] = ((val)>>24)&255), \
		 ((p)[off++] = ((val)>>16)&255), \
		 ((p)[off++] = ((val)>>8)&255), \
		 ((p)[off++] = ((val)>>0)&255))

static int appendString(char *p, int off, const char *s)
{
	int len;
	if (s == null) {
		len = 0;
		s = "";
	} else {
		len = strlen(s);
	}
	appendInt(p, off, len);
	strncpy(p+off, s, len);
	return off + len;
}

static int
sendRequest(poolEntry *p, jrun_request *r, const char *target)
{
	int i, len = 0, alen=0;
	char *h;
	const char *key;
	char *vars[VARIABLE_COUNT], **hdrs;
	int off;

	/*
	 * first calculate how big a buffer we need
	 */
	char **attrNames = r->getAttributeNames(r);
	if (attrNames != null) {
		for (i=0; attrNames[i] != null; i++) {
			int attrLen = r->getAttributeSize(r, attrNames[i]);
			alen += 12 + strlen(attrNames[i]) + attrLen;
			r->trace(r, "   %.200s: %d bytes", attrNames[i], attrLen);
		}
	}

	r->trace(r, "PROXY_SERVICE_30 -> [%d]", p->sock);

	/* first get all the variable values, adding up length */
	len += 4;
	for (i=0; i < VARIABLE_COUNT; i++) {
		key = VARIABLE_NAMES[i].var;
		vars[i] = r->getHeader(r, key);
		len += 4 + 4 + (vars[i] ? strlen(vars[i]) : 0);
	}
	/* room for URI */
	len += 4 + 4 + strlen(r->uri);

	if (target != null) {
		/* make room for PROXY_SET_JRUN_TARGET */
		len += 4 + 4 + strlen(target);
	}

	/* get all the headers.  getHeaderNames must not return NULL */
	for (i=0; r->hdrNames[i] != NULL; i++);
	hdrs = r->malloc(r, i * sizeof (char *));
	for (i=0; r->hdrNames[i] != NULL; i++) {
		key = r->hdrNames[i];
		hdrs[i] = r->getHeader(r, key);
		len += 4 + 4 + strlen(key) + 4 + (hdrs[i] ? strlen(hdrs[i]) : 0);
	}

	/* add room for the trailing PROXY_SERVICE_30 command */
	len += 4;

	/*
	 * now allocate memory and fill in the packet
	 */

	/* grab a chunk of memory to serialize the headers */
	h = r->malloc(r, alen + len + 4 + 1);
	if (h == null) {
		r->log(r, "malloc(%d) failed in sendAllHeaders -> [%d]",
			   alen + len + 1 + 2 * sizeof(int), p->sock);
		for (i=0; i < VARIABLE_COUNT; i++) {
			if (vars[i] != null)
				r->free(r, vars[i]);
		}
		for (i=0; r->hdrNames[i] != NULL; i++) {
			if (hdrs[i] != null) {
				r->free(r, hdrs[i]);
			}
		}
		return -1;
	}

	off = 0;

	appendInt(h, off, PROXY_SERVICE_30); /* send initial command */
	appendInt(h, off, alen + len - 4); /* total # bytes to follow */

	if (attrNames != null) {
		r->trace(r, "PROXY_SET_ATTRIBUTE -> [%d]", p->sock);
		for (i=0; attrNames[i] != null; i++) {
			int attrLen = r->getAttributeSize(r, attrNames[i]);
			int nameLen = strlen(attrNames[i]);
			appendInt(h, off, PROXY_SET_ATTRIBUTE);
			off = appendString(h, off, attrNames[i]);
			appendInt(h, off, attrLen);
			r->getAttribute(r, attrNames[i], h+off);
			off += attrLen;
			r->trace(r, "   %.200s %d bytes", attrNames[i], attrLen);
		}
	}

	/* send the variables */
	for (i = 0; i < VARIABLE_COUNT; i++) {
		int cmd = VARIABLE_NAMES[i].cmd;
		appendInt(h, off, cmd);
		off = appendString(h, off, vars[i]);
		if (vars[i] != null)  {
			r->free(r, vars[i]);
		}
	}

	/* send the request uri */
	appendInt(h, off, PROXY_SET_REQUEST_URI);
	off = appendString(h, off, r->uri);

	if (target != null) {
		/* send name of servlet to invoke */
		appendInt(h, off, PROXY_SET_JRUN_TARGET);
		off = appendString(h, off, target);
	}

	/* add the headers */
	for (i = 0; r->hdrNames[i] != NULL; i++) {
		key = r->hdrNames[i];
		appendInt(h, off, PROXY_SET_HEADER);
		off = appendString(h, off, key);
		off = appendString(h, off, hdrs[i]);
		if(hdrs[i] != null) { 
			r->free(r, hdrs[i]);
		}
	}
	r->free(r, hdrs);

	appendInt(h, off, PROXY_SERVICE_30); /* terminate packet */

	if (off != (int)(alen + len + 4)) {
		r->log(r, "sendAllHeaders end %d should be %d", off, alen+len+2*sizeof(int));
	}

	i = sendBytes(r, h, off, p->sock);
	r->free(r, h);
	return (i == off) ? 1 : -1;
}

/*
   502 Bad Gateway

   The server, while acting as a gateway or proxy, received an invalid
   response from the upstream server it accessed in attempting to
   fulfill the request.

   503 Service Unavailable

   The server is currently unable to handle the request due to a
   temporary overloading or maintenance of the server. The implication
   is that this is a temporary condition which will be alleviated
   after some delay.
*/

/**
 * proxyerror() gets called on any socket related proxy error.  It either produces
 * an error page, or redirects the browser to a user-defined error page.
*/

static void proxyerror(jrun_proxy *pxy, jrun_request *r)
{
	char *errorMsg, *statusMsg;
	int statusCode = 503;
	int error = r->error;

	switch (error) {
		/* jrun dropped our socket, usually due to a restart */
	case ECONNRESET:
	case ESHUTDOWN:
		statusMsg = "JRun unexpectedly closed connection";
		errorMsg = "<H2>JRun closed connection</H2>\r\n";
		break;

		/* proxy connect error */
	case ECONNREFUSED:
	case ETIMEDOUT:
		statusMsg = "Could not connect to JRun";
		errorMsg = "<H2>Could not connect to JRun Server</H2>\r\n";
		break;

		/* jrun rejected our connection because it was too busy */
	case EJRUN_BUSY:
		statusMsg = "Too many concurrent requests";
		errorMsg = "<H2>Too many concurrent requests</H2>jcp.endpoint.main.max.threads exceeded\r\n";
		break;
	
		/* general error while doing proxy protocol. this should never happen */
	default:
		statusMsg = "JRun Connector protocol error";
		errorMsg = "<H2>JRun Connector Protocol Error</H2>\r\n";
		break;
	}

	r->trace(r, "returning error page for %s", describeError(error));
	r->setStatus(r, statusCode, statusMsg);
	r->addHeader(r, "content-type", "text/html");
	r->writeClient(r, errorMsg, strlen(errorMsg));
	r->committed = TRUE;
	free_jrun_request(r);
}

static int sendPostData(jrun_request *r, int clen, poolEntry *p)
{
	int actual;
	char *mybuf;
	int *intbuf;
	
	r->trace(r, "   getPostData content-length: %d", clen);
	mybuf = (char *) r->malloc(r, 8+(clen < POST_CHUNK_SIZE ? clen : POST_CHUNK_SIZE));
	if (mybuf == null) /* out  of memory */
		return 0;
	intbuf = (int *) mybuf;

	actual = clen < POST_CHUNK_SIZE ? clen : POST_CHUNK_SIZE;
	actual = r->getPostData(r, mybuf+8, actual);
	r->trace(r, "getPostData returned %d", actual);
	intbuf[0] = htonl(PROXY_PUT_DATA);
	intbuf[1] = htonl(actual);
	actual = (actual > 0) ? actual+8 : 8;
	if (actual < 0) {
		r->log(r, "sendPostData: error reading from client");
	}
	if (sendBytes(r, mybuf, actual, p->sock) == -1) {
		r->log(r, "sendPostData: error sending to proxy");
		r->free(r, mybuf);
		return 0;
	} else {
		r->free(r, mybuf);
		return 1;
	}
}


/* nextHeader is used to parse through a string of headers.  Each header
must be in the format: <name>: <value>\r\n, and the block should
be terminated with an additional \r\n.  

  Return value: pointer to beginning of next header, or null if no more
  headers were found.
  return-by-reference: pname = pointer to header name, null terminated.
  return-by-reference: pvalue = pointer to header value, null terminated.

  WARNING: this function puts nulls into the header block as it is parsing.
  DO NOT pass a static string!  linkers often allocate static stings in 
  read-only memory.
  */

static char *nextHeader(char *headers,
			   char **pname, char **pvalue)
{
		char *eol = null, *value;

		value = strstr(headers, ": ");
		if (value != null) {
			*value = '\0';
			value += 2;
			eol = strstr(value, "\r\n");
			if (eol != null) {
				while (eol[2] == ' ' || eol[2] == '\t') { /* continuation */
					eol = strstr(eol+2, "\r\n");
				}
				*eol = '\0';
				eol += 2;
				*pname = headers;
				*pvalue = value;
			}
		}
		return eol;
}

/* 
 * handleRequest
 *
 * return 0 == success, -1 == failure
 *
 */
static int
handleJrppRequest(jrun_proxy *pxy, jrun_request *r, const char *target)
{
	sockPool *spool = (sockPool *) pxy->obj1;
	poolEntry *p = null;
	int command;
	
	r->locked = 0;
	r->trace(r, "%s handleJrppRequest()", pxy->name);

	/* Get a connection to the JRun proxy server for handling the request */
retryRequest:
	if((p = getJRunConnectionFromPool(pxy, r)) == null ){
		proxyerror(pxy, r);
		return 0;
	}

	/* Send initial command and headers */
	if(sendRequest(p, r, target) == -1 ) {
		int pooled = (p->next != null);
		r->trace(r, "Couldn't send request to [%d]", p->sock);
		destroyJRunConnection(r,  pxy, p);
		if (pooled) {
			r->trace(r, "   was pooled, trying another connection");
			goto retryRequest;
		}
		proxyerror(pxy, r);
		return 0;
	}

	/* set init mode */
	command = PROXY_UNKNOWN;

	/* wait until we're told to quit */
	while( command != PROXY_FINISH && command > 0)
	{
		time_t before = time(0);
		/* get command */
		if( readFully(r, &command, sizeof(int), p->sock) == -1 ) {
			int pooled = (p->next != null);
			if (pooled && command == PROXY_UNKNOWN) {
				/* socket was pooled but had a problem -- try another socket */
				r->trace(r, "   read error - was pooled. trying again");
				destroyJRunConnection(r, pxy, p);
				goto retryRequest;
			}
			if (command == PROXY_UNKNOWN) {
				/* socket was closed */
				r->log(r, "[%d] dropped.", p->sock);
				r->error = ESHUTDOWN;
			} else {
				r->log(r, "command error: %s <- [%d]", describeError(r->error), p->sock);
			}
			break;
		}

		/* convert to proper endian */
		command = ntohl(command);

		switch(command) {

			case PROXY_BUSY:
				r->log(r, "PROXY_BUSY <- [%d]", p->sock);
				r->error = EJRUN_BUSY;
				command = -1;
				break;

			case PROXY_GET_HEADER:
				{
					char *name;
					r->trace(r, "PROXY_GET_HEADER <- [%d]", p->sock);
					name = readData(p->sock, r, null);
					if (name == null) {
						command = -1; /* r->error contains error code */
					} else {
						char *value = r->getHeader(r, name);
						if (value == null) {
							int i;
							for (i=0; name[i] != '\0'; i++) {
								if (name[i] == '-') name[i] = '_';
								else if (name[i] == '_') name[i] = '-';
							}
							value = r->getHeader(r, name);
						}
						if (writeString(r, p->sock, value) == -1)
							command = -1; /* r->error contains error code */
						if (value != null)
							r->free(r, value);
						r->free(r, name);
					}
				}
				break;

			case PROXY_GET_REALPATH:
				{
					char *path;
					r->trace(r, "PROXY_GET_REALPATH <- [%d]", p->sock);
					path = readData(p->sock, r, null);
					if (path == null) {
						command = -1;
					} else {
						char *value = r->getRealPath(r, path);
						if (writeString(r, p->sock, value) == -1)
							command = -1;
						if (value != null) 
							r->free(r, value);
						r->free(r, path);
					}
				}
				break;

			/* read a given number of bytes without stopping */
			case PROXY_GET_DATA:
				{
					int len;
					r->trace(r, "PROXY_GET_DATA <- [%d]", p->sock);
					if (readFully(r, &len, sizeof(int), p->sock) == -1) {
						r->log(r, "getData: can't get length");
						command = -1;
					} else {
						len = ntohl(len);
					}
					if (!sendPostData(r, len, p)) {
						command = -1;
					}
					break;
				}

			case PROXY_PUT_STATUS:
				{
					char *statusLine;
					int statusCode;
					r->trace(r, "PROXY_PUT_STATUS <- [%d]", p->sock);
					/* read the status code, then the status string len */
					if(readFully(r, &statusCode, sizeof(int), p->sock) == -1 ||
						(statusLine = readData(p->sock, r, null)) == null) {
						command = -1;
					} else {
						statusCode = htonl(statusCode);
						r->trace(r, "   %d %.80s", statusCode, statusLine);
						r->setStatus(r, statusCode,	statusLine);
						r->free(r, statusLine);
					}
				}
				break;

			case PROXY_PUT_HEADER:
				{
					char *headers, *next, *name, *value;
					r->trace(r, "PROXY_PUT_HEADER <- [%d]", p->sock);
					headers = readData(p->sock, r, null);
					if (headers == null) {
						command = -1;
					} else {
						next = headers;
						while ((next = nextHeader(next, &name, &value)) != null) {
							r->addHeader(r, name, value);
						}
						r->free(r, headers);
					}
				}
				break;

			case PROXY_PUT_HEADER_VALUE:
				{
					char *name, *value;
					r->trace(r, "PROXY_PUT_HEADER_VALUE <- [%d]", p->sock);
					name = readData(p->sock, r, null);
					if (name == null) {
						command = -1;
					} else {
						value = readData(p->sock, r, null);
						if (value == null) {
							r->free(r, name);
							command = -1;
						} else {
							r->addHeader(r, name, value);
							r->free(r, value);
							r->free(r, name);
						}
					}
				}
				break;
			
			case PROXY_SET_ATTRIBUTE:
				{
					int len;
					char *name, *attribute;
					r->trace(r, "PROXY_SET_ATTRIBUTE <- [%d]", p->sock);
					name = readData(p->sock, r, null);
					if (name == null) {
						command = -1;
						r->trace(r, "   socket error");
						break;
					}
					attribute = readData(p->sock, r, &len);
					if (attribute == null) {
						command = -1;
						r->free(r, name);
						r->trace(r, "   socket error");
						break;
					}
					r->setAttribute(r, name, attribute, len);
					r->trace(r, "   %.200s %d bytes", name, len);
					r->free(r, attribute);
					r->free(r, name);
				}
				break;

			case PROXY_PUT_CLIENT:
				{
					int len;
					char *val;
					r->trace(r, "PROXY_PUT_CLIENT <- [%d]", p->sock);
					val = readData(p->sock, r, &len);
					if (val == null ) {
						command = -1;
					} else {
						if( r->writeClient(r, val, len) == -1 )
							command = -1;
						r->committed = TRUE;
						r->free(r, val);
					}
				}
				break;
			
			case PROXY_FINISH:
				{
					r->trace(r, "PROXY_FINISH <- [%d]", p->sock);
					r->writeClient(r, null, 0); /* flush */
				}
				break;

			default:
				r->log(r, "handleJrppRequest: bad command: %d <- [%d]", command, p->sock);
				r->error = EJRUN_PROTOCOL_ERROR;
				command = -1;
				break;
		}
	}

	if (command != PROXY_FINISH) {
		r->trace(r, "handleJrppRequest aborting");
		if (r->error == ENOBUFS) {
			/* internal error or memory error */
			destroyJRunConnection(r, pxy, p);
			return -1;
		} else {
			destroyJRunConnection(r, pxy, p);
			proxyerror(pxy, r);
			return 0;
		}
	} else {
		returnJRunConnectionToPool(r, pxy, p);
		return 0;
	}
}

static poolEntry *
sendCommand(jrun_proxy *pxy, jrun_request *r, const char *command, int len, int *response)
{
	poolEntry *p;
retry:
	if ((p = getJRunConnectionFromPool(pxy, r)) == null) {
		return null;
	}
	if (sendBytes(r, command, len, p->sock) == -1) {
		int pooled = (p->next != null);
		r->trace(r, "   Couldn't send command to [%d]", p->sock);
		destroyJRunConnection(r, pxy, p);
		if (pooled) {
			r->trace(r, "   was pooled, trying again");
			goto retry;
		} else {
			return null;
		}
	}
	if (readFully(r, response, sizeof(int), p->sock) == -1) {
		if (p->next != NULL) {
			/* try another socket */
			r->trace(r, "   was pooled, trying again");
			destroyJRunConnection(r, pxy, p);
			goto retry;
		}
		/* socket was closed. other side is busy */
		r->log(r, "JRun is busy. [%d] dropped while sending command", p->sock);
		return null;
	}
	*response = ntohl(*response);
	return p;
}


static int prefixlen(const char *map)
{
	const char *s, *e;
	if (strncmp(map, "//", 2) == 0) {
		if ((s = strchr(map, '/')) == null)
			s = strchr(map, '\0');
	} else {
		s = map;
	}
	if ((e = strchr(s, '*')) == null) {
		e = strchr(s, '\0');
	}
	return e-s;
}

static int extlen(const char *map)
{
	const char *e;
	if ((e = strchr(map, '*')) == null) {
		return 0;
	} else {
		return strlen(e+1);
	}
}

/*
 * return 1 if a is longer than b, 0 otherwise
 */
static int longerThan(propEntry *a, propEntry *b)
{
	int ahost = (0 == strncmp(a->name, "//", 2));
	int bhost = (0 == strncmp(b->name, "//", 2));
	int alen, blen;
	if (ahost && !bhost) return 1;
	if (!ahost && bhost) return 0;
	alen = prefixlen(a->name);
	blen = prefixlen(b->name);
	if (alen != blen) {
		return alen > blen;
	}
	alen = extlen(a->name);
	blen = extlen(b->name);
	return alen > blen;
}
		
/**
 * sort an array of mappings, in place, so the most-specific match is first.
 * this is an insertion sort, so it is quick when the array is already sorted.
 */
static void sortMappings(propTable *mappings)
{
	int i, j;
	propEntry *maps = mappings->entry;
	/* clean them up first */
	for (i=0; i < mappings->count; i++) {
		char *m = mappings->entry[i].name;
		char *e = strchr(m, '*');
		if (e == null) e = strchr(m, '\0');
		if (e > m && e[-1] == '/') {
			for (; e[-1] != '\0'; e++) {
				e[-1] = e[0];
			}
		}
	}
	/* now sort them */
	for (i=1; i < mappings->count; i++) {
		propEntry m = maps[i];
		for (j=i; j > 0 && longerThan(&m, &maps[j-1]); --j) {
			maps[j] = maps[j-1];
		}
		maps[j] = m;
	}
}

static propTable *fetchProps(jrun_proxy *pxy, jrun_request *r, const char *file)
{
	int len, command;
	char *cmd;
	sockPool *spool = (sockPool *) pxy->obj1;
	poolEntry *p = null;
	propTable *pTable = null;

	r->trace(r, "%s fetching %.250s", pxy->name, file);
	len = strlen(file);
	cmd = (char *) r->malloc(r, 2 * sizeof(int) + len);
	*((int *)cmd) = htonl(PROXY_GET_PROPS);
	appendString(cmd, 4, file); 

	if ((p = sendCommand(pxy, r, cmd, len + 8, &command)) == null) {
		r->free(r, cmd);
		return null;
	} else {
		r->free(r, cmd);
	}

	/* get the response */
	if (command == PROXY_PUT_PROPS) {
		char *rules;
		r->trace(r, "PROXY_PUT_PROPS <- [%d]", p->sock);
		rules = readData(p->sock, r, null);
		if (rules != null) {
			pTable = parseProperties(r, rules);
			r->free(r, rules);
			sortMappings(pTable);
		}
		returnJRunConnectionToPool(r, pxy, p);
	} else if (command == PROXY_BUSY) {
		r->log(r, "JRun unable to fetch %.250s <- [%d]", file, p->sock);
		destroyJRunConnection(r, pxy, p);
	} else {
		r->log(r, "Unexpected command %d <- [%d] while fetching %.250s", command, p->sock, file);
		destroyJRunConnection(r, pxy, p);
	}
	return pTable ;
}

static jrun_proxy *
mapRequestCheckAll(jrun_proxy *pxy, jrun_request *r)
{
	jrun_proxy *p, *target = null;

	for (p = pxy->pxylist; p != null && target == null; p = p->next) {
		target = p->filterRequest(p, r);
	}
	return target;
}

static void refreshMappings(struct jrun_proxy *pxy, jrun_request *r)
{
	propTable *mappings, *oldMappings;
	/* try to load rules */

	time_t now = time(0);
	if (now > pxy->next_mapcheck_time) {
		startSync(r, pxy);
		if (now <= pxy->next_mapcheck_time) {
			/* somebody else beat us to it */
			stopSync(r, pxy);
			return;
		} else {
			/* set the next check time to prevent thundering herd of checks */
			pxy->next_mapcheck_time = now + pxy->mapcheck_interval;
			stopSync(r, pxy);
		}

		r->trace(r, "refreshing mappings for %s", pxy->name);
		mappings = fetchProps(pxy, r, "servlet-mappings");
		if (mappings != null) {
			/* install the maps we just fetched */
			r->trace(r, "installing new mappings for %s", pxy->name);
			startSync(r, pxy);
			oldMappings = pxy->mappings;
			pxy->mappings = mappings;
			stopSync(r, pxy);
			if (oldMappings != NULL) {
				freeProperties(r, oldMappings);
			}
		} else {
			/* could not get mappings -- keep using old ones */
			r->trace(r, "could not get mappings, checking again in %d seconds", pxy->mapcheck_interval);
		}
	}
}

/*
 * mapRequestCheck - this is the guts of JRun's filtering.
   We use URI when doing prefix mapping, use path when doing suffix mapping.
   The reason for this is that the path can be different from the URI when 
   the web server is serving a welcome file.
   best example is this:
          URI   = /
		  path  = <docroot>/index.jsp

   @returns the proxy that will handle the request
*/

static jrun_proxy *
mapRequestCheck(struct jrun_proxy *pxy, jrun_request *r)
{
	int i, matchLen = -1, pathLen, uriLen;
	char *reqHost;
	char *path = r->path;
	char *uri = r->uri;
	const char *map = null;

	refreshMappings(pxy, r);
	startSync(r, pxy);
	if (pxy->mappings == null) {
		stopSync(r, pxy);
		r->log(r, "local.properties not configured.  Skipping request");
		return null;
	}

	if (path == NULL) {
		path = uri;
	}
	pathLen = strlen(path);
	uriLen = strlen(uri);
	reqHost = r->getHeader(r, "HOST");
	r->trace(r, "filtering %.200s (%.200s) HOST=%s", uri, path, reqHost ? reqHost : "null");
	for( i=0; i < pxy->mappings->count && map == null;i++) {
		const char *prefix, *ext, *host = null;
		int prefixLen, extLen, hostLen;
		map = pxy->mappings->entry[i].name;

		/* check for //host/ at the beginning of a rule to 
		   indicate a multi-homing rule */
		if (strncmp(map,"//",2) == 0) {
			if ((prefix = strchr(map+2, '/')) == null) {
				prefix = strchr(map+2,'\0');
			}
			host = map+2;
			hostLen = (prefix - map) - 2;
		} else {
			/* no host prefix found */
			prefix = map;
			host = null;
			hostLen = 0;
		}
		/* now prefix points to remainder of rule, prefixLen not set */
		if ((ext = strchr(prefix,'*')) != null) {
			prefixLen = ext-prefix;
			++ext;
		} else {
			ext = strchr(prefix, '\0');
			prefixLen = strlen(prefix);
		}
		extLen = strlen(ext);

		if ((host == null || reqHost != null &&
							 strncasecmp(host, reqHost, hostLen) == 0 &&
							 reqHost[hostLen] == '\0') &&
			strncmp(uri, prefix, prefixLen) == 0 &&
			(uriLen >= extLen && strcasecmp(&uri[uriLen-extLen], ext) == 0 ||
			 pathLen >= extLen && strcasecmp(&path[pathLen-extLen], ext) == 0)) {

			if (prefixLen < uriLen) {
				char delim = uri[prefixLen];
				if (delim != '/' && delim != '&' && delim != '#') {
					continue;
				}
			}
			r->trace(r, "   matched %s", map);
		} else {
			map = null;
		}
	}
	
	stopSync(r, pxy);
	if (reqHost != null) r->free(r, reqHost);
	if (map != null) {
		return pxy;
	} else {
		return null;
	}
}

/**
 * loadMappings - read a local.properties file and extract the mapping rules in
 * the form "jrun.rules.<pattern>=<servlet>".  Associate these mappings with the
 * server given.
 */
static void loadMappings(jrun_proxy *pxy, jrun_request *r, propTable *props)
{
	propEntry *e, *e2;
	char *newName;
	int i, j;
	char buf[512];

	if (pxy->mappings != null) {
		freeProperties(r, pxy->mappings);
	}

	pxy->mappings = r->perm_malloc(r, sizeof(propTable));
	pxy->mappings->count = 0;

	for (i=0; i < props->count; i++) {
		e = &props->entry[i];
		if (strncmp(e->name, "webapp.mapping.", 15) == 0) {
			const char *useRoot;
			strcpy(buf, e->value);
			strcat(buf, ".use-webserver-root");
			useRoot = getPropertyValue(props, buf);
			if (useRoot != null && strcasecmp(useRoot, "true") == 0) {
				for (j=0; j < props->count; j++) {
					e2 = &props->entry[j];
					if (strncmp(e2->name, "webapp.servlet-mapping.", 23) == 0) {
						newName = r->perm_malloc(r,
							strlen(e->name)-15+strlen(e2->name)-23+1);
						strcpy(newName, e->name+15);
						strcat(newName, e2->name+23);
						addProperty(pxy->mappings, newName, e->value);
					}
				}
			} else {
				newName = r->perm_malloc(r, strlen(e->name)+1-15);
				strcpy(newName, e->name+15);
				addProperty(pxy->mappings, newName, e->value);
			}
		} else if (e->name[0] == '/' || e->name[0] == '*') {
			/* old-style rules.properties */
			addProperty(pxy->mappings, e->name, e->value);
			e->name = null;
		}
                e->value = null;
	}
	sortMappings(pxy->mappings);
	r->trace(r, "loadMappings - after sort");
	for (i=0; i < pxy->mappings->count; i++) {
		r->trace(r, "   %s=%s", pxy->mappings->entry[i].name,
			pxy->mappings->entry[i].value);
	}
}

/**
 * destroy and free the entire proxy data structure.  Normally called
 * at shutdown time.
 */
static void
destroyInetProxy(jrun_proxy *pxy, jrun_request *r)
{
	sockPool *spool = (sockPool *) pxy->obj1;

	while (spool->pool != null) {
		poolEntry *p = spool->pool;
		spool->pool = p->next;
		r->trace(r, "%s closing socket %d", pxy->name, p->sock);
		destroyJRunConnection(r, pxy, p);
	}
	r->free(r, spool);
	r->trace(r, "%s destroyed.", pxy->name);
}

static void destroyProxy(jrun_proxy *pxy, jrun_request *r)
{
	jrun_proxy *p;

	for (p = pxy->pxylist; p != null; p = p->next) {
		p->destroy(p, r);
	}
}

/* initialize a JRPP (aka "inet") proxy.  Returns 0 = success, -1 = failure */
static int initInetProxy(jrun_proxy *pxy, jrun_request *r,
						 const char *jvmname, const char *localProps,
						 propTable *global)
{
	char hostbuf[20]; /* 20 is enough for any IP address (xxx.xxx.xxx.xxx) */
	const char *e, *host;
	int port;
	propTable *local;

	r->trace(r, "initInetProxy");
	initSync(r, pxy); /* initialize pool lock */

	if ((local = loadProperties(r, localProps)) == null) {
		return -1;
	}
	if (atoi(jvmname) == 0) {
		/* get ip + port from property file */
		const char *services = getPropertyValue2(local, global, "jrun.services");
		if (services != null && (e = strstr(services,"jcp")) != null && !isalnum(e[3])) {
			const char *strPort = getPropertyValue2(local, global, "jcp.endpoint.main.port");
			if (strPort != null) port = atoi(strPort);
			host = getPropertyValue2(local, global, "jcp.endpoint.main.bindaddress");
			if (host != null && strcmp(host, "*") == 0) {
				host = JRUN_PROXY_HOST;
			}
		} else {
			/* jcp service not activated */
			freeProperties(r, local);
			r->trace(r, "%s does not have an active JCP service -- skipping", jvmname);
			return -1;
		}
	} else {
		/* set up server ip/port info */
		if ((e = strchr(jvmname, ':')) != null) {
			strncpy(hostbuf, jvmname, e-jvmname);
			hostbuf[e-jvmname] = '\0';
			host = hostbuf;
			port = atoi(e+1);
		} else {
			r->log(r, "Invalid server %.200s. "
				   "Please use <ip>:<port> notation", jvmname);
			freeProperties(r, local);
			return -1;
		}
	}
	if (e == null || host == null || inet_addr(host) == -1) {
		r->log(r, "Invalid server %.200s, check settings and local.properties", jvmname);
		freeProperties(r, local);
		return -1;
	}
	if (port < 1 || port > 65535) {
		r->log(r, "Invalid proxy port: %d.  Port must be between 1 and 65535");
		freeProperties(r, local);
		return -1;
	}
	strcpy(pxy->name, jvmname);
	pxy->obj1 = initPool(r, pxy, host, port);
	if (pxy->obj1 == null) {
		freeProperties(r, local);
		return -1;
	}

	pxy->mapcheck_interval = 10; /* seconds */
	pxy->next_mapcheck_time = 0;

	/* set up callbacks */
	pxy->filterRequest = mapRequestCheck;
	pxy->handleRequest = handleJrppRequest;
	pxy->destroy = destroyInetProxy;
	r->trace(r, "JRun Connector Proxy configured for proxy host %.80s, port %d", host, port);
	loadMappings(pxy, r, local);
	freeProperties(r, local);
	return 0; /* success */
}

/**
 * process the given request.  If targetServlet is not null, the named 
 * servlet will be invoked on the server, bypassing the server's mapping
 * rules.  This is similar to calling getNamedDispatcher.forward()
 */
static int
handleRequest(jrun_proxy *pxy, jrun_request *r, const char *targetServlet)
{
	jrun_proxy *p;

	p = pxy->filterRequest(pxy, r);
	if (p == pxy || p == null) {
		r->log(r, "ERROR: no server matched target URL %.500s", r->uri);
		return -1;
	} else {
		return p->handleRequest(p, r, targetServlet);
	}
}


/**
add a list of servers and/or rules. 
*/
static int addJvms(jrun_proxy *pxy, jrun_request *r,
				   const char *jrunroot, const char *jvmlist)
{
	jrun_proxy *inetProxy;
	char path[512];
	propTable *jvms, *global;
	int i;

	sprintf(path, "%s/lib/jvms.properties", jrunroot);
	if ((jvms = loadProperties(r, path)) == null) {
		return -1;
	}
	sprintf(path, "%s/lib/global.properties", jrunroot);
	if ((global = loadProperties(r, path)) == null) {
		freeProperties(r, jvms);
		return -1;
	}

	for (i=0; i < jvms->count; i++) {
		if (jvmlist != null && strstr(jvmlist, jvms->entry[i].name)==null) {
			/* this jvm is not in jvm list -- skip it */
			r->trace(r, "%s is not in jvm list -- skipping", jvms->entry[i].name);
			continue;
		}
		inetProxy = r->perm_malloc(r, sizeof(jrun_proxy));
		if (inetProxy == null) {
			freeProperties(r, jvms);
			freeProperties(r, global);
			return -1;
		}
		memset(inetProxy, 0, sizeof(*inetProxy));
		sprintf(path, "%s/local.properties", jvms->entry[i].value);
		if (initInetProxy(inetProxy, r, jvms->entry[i].name, path, global) != 0) {
			r->perm_free(r, inetProxy);
			r->log(r, "could not initialize %s", jvms->entry[i].name);
		} else {
			inetProxy->next = pxy->pxylist;
			pxy->pxylist = inetProxy;
		}
	}
	freeProperties(r, global);
	freeProperties(r, jvms);
	return 0;
}

static int addServer(struct jrun_proxy *pxy, jrun_request *r, const char *server)
{
	const char *e;
	char address[100];
	jrun_proxy *inetProxy;

	e = strchr(server, ';');
	if (e == null) {
		r->log(r, "invalid server spec: %s", server);
		return -1;
	}
	strncpy(address, server, e-server);
	address[e-server] = '\0';
	for (++e; isspace(*e); ++e);

	inetProxy = r->perm_malloc(r, sizeof(jrun_proxy));
	if (inetProxy == null) {
		return -1;
	}
	memset(inetProxy, 0, sizeof(*inetProxy));
	if (initInetProxy(inetProxy, r, address, e, null) != 0) {
		r->perm_free(r, inetProxy);
		r->log(r, "could not initialize %s", address);
	} else {
		inetProxy->next = pxy->pxylist;
		pxy->pxylist = inetProxy;
	}
	return 0;
}

void initProxy(jrun_proxy *pxy, jrun_request *r, const char *jrunroot, const char *jvms)
{
	memset(pxy, 0, sizeof(*pxy));
	initSync(r, pxy);
	pxy->filterRequest = mapRequestCheckAll;
	pxy->handleRequest = handleRequest;
	pxy->destroy = destroyProxy;
	pxy->addServer = addServer;
	strcpy(pxy->name, "default");
	if (jrunroot != null) {
		addJvms(pxy, r, jrunroot, jvms);
	}
}
