/*
Copyright (c) 1997-2000 Allaire.  All Rights Reserved.

DO NOT REDISTRIBUTE THIS SOFTWARE IN ANY WAY WITHOUT THE EXPRESSED
WRITTEN PERMISSION OF ALLAIRE.
*/

#ifndef JRUN_PROXY_H
#define JRUN_PROXY_H

#if defined __cplusplus
extern "C" {
#endif

/* headers for locking calls */
#if defined(XP_WIN32) || defined(XP_NETWARE) || defined(XP_UNIX)
#include <base/systems.h>
#include <base/crit.h>
#elif defined(WIN32)
#include <windows.h>
#endif

/* headers for socket calls */
#ifdef WIN32

#include <winsock.h>
#include <time.h>
#undef EINTR
#define EINTR WSAEINTR
#define ETIMEDOUT WSAETIMEDOUT
#define ECONNREFUSED WSAECONNREFUSED
#define EWOULDBLOCK WSAEWOULDBLOCK
#define ESHUTDOWN WSAESHUTDOWN
#define ECONNRESET WSAECONNRESET
#define ENOBUFS WSAENOBUFS
#undef strcasecmp
#undef strncasecmp
#define strcasecmp _strcmpi
#define strncasecmp _strnicmp
#define FILE_SEPARATOR '\\'

#else /* !WIN32 */
#define FILE_SEPARATOR '/'
typedef int SOCKET;
#define WSAGetLastError() (errno)

#ifdef XP_NETWARE
#include <stdio.h>
#include <nwtypes.h>
#include <sys/bsdskt.h>
#define TCP_NODELAY  1
#define strcasecmp strcmpi
#define strncasecmp strnicmp
#endif /* XP_NETWARE */

#include <sys/types.h>
#include <sys/socket.h>
#include <time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#ifndef XP_NETWARE
#include <netinet/tcp.h>
#endif

#include <errno.h>
#include <unistd.h>
#define TRUE 1
#define FALSE 0

#endif /* !WIN32 */

#ifndef XP_NETWARE
#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <stdlib.h>
#endif /* !XP_NETWARE */

/* 

This header contains interface structures used to abstract
the connector-side of a JRun plugin (jrun_request) and the
proxy side (jrun_proxy).  Each contains functions and data
needed by implementations of these interfaces. 

*/


/* connector-side interface.  Use this structure to
   encapsulate the plugin API.  These routines are
   called by the proxy.

   This data structure contains methods for accessing static
   (global) and dynamic (per-request) web server resources.
   This includes web stuff, network I/O, memory allocation,
   and error logging.
*/

typedef struct jrun_request {

	/* generic pointers for server-specific objects */
	void * context;
	void * connection;
	void * request;

	/* printable handle for logging */
	char stringRep[64];

	int error; /* used to hold error code for last operation */
	int noaction, /* set this so we know to ignore writeclient calls */
		locked, /* debug flag to detect double-locks  */
		shouldRead, /* ok to block reading post data? */
		shouldReadMore,
		committed; /* has response been committed yet? */
	int offset; /* used by isapi filter to keep track of input position */

	/* http status/reason string */
	char *reason; 

	/* these three members are used by ISAPI to rebuild header string */
	char *headers;
	char *headersEnd;
	int headersSize;

	/* used by web-server side to pass to jrun side. these MUST be defined */
	char **hdrNames;
	char *uri, *path;

	/* servicing requests */
	char *(*getHeader)(struct jrun_request *, const char *headerName);
	char *(*getRealPath)(struct jrun_request *, const char *virtualPath);
	int (*getPostData)(struct jrun_request *, char *buf, int len);
	void (*setStatus)(struct jrun_request *, int status, const char *reason);
	void (*addHeader)(struct jrun_request *, const char *header, const char *value);
	int (*writeClient)(struct jrun_request *, const char *data, int len);

	/* attributes, new for JRun Matrix */
	void (*getAttribute)(struct jrun_request *, const char *name, char *buf);
	int (*getAttributeSize)(struct jrun_request *, const char *name);
	char **(*getAttributeNames)(struct jrun_request *);
	void (*setAttribute)(struct jrun_request *, const char *name, const char *buf, int len);

	/* memory allocation */
	void *(*malloc)(struct jrun_request *, int nbytes);
	char *(*jstrdup)(struct jrun_request *, const char *s);
	void (*free)(struct jrun_request *, void *);
	void *(*perm_malloc)(struct jrun_request *, int nbytes);
	void (*perm_free)(struct jrun_request *, void *);

	/* logging */
	void (*log)(struct jrun_request *, const char *fmt, ...);
	void (*trace)(struct jrun_request *, const char *fmt, ...);

	/* network api */
	int (*connect)(struct jrun_request *, struct sockaddr_in *addr);
	void (*close)(struct jrun_request *, int fd);
	int (*send)(struct jrun_request *, int fd, const char *buf, int len);
	int (*recv)(struct jrun_request *, int fd, char *buf, int len);

} jrun_request;

/* proxy interface.  Used to encapsulate access to JRun.
   These routines are called by the native connector.

   This data structure represent's the web-server-specific's view
   of the servlet engine.  There can and will be diferent implementations
   of this interface.  For example:
       - One jrun_proxy structure can keep track of multiple
	     other jrun_proxy objects internally, and choose between them
		 based on: 
		     - load
			 - application prefix
			 - session
			 - other
	   - one jrun_proxy object for each remote server (JRPP)
	   - one jrun_proxy object for each internal server (JNI) -- future
	   - etc.
*/

typedef struct jrun_proxy {
#if defined(XP_WIN32) || defined(XP_UNIX) || defined(XP_NETWARE)
	CRITICAL lock;
#elif defined(WIN32)
	CRITICAL_SECTION lock;
#endif /* !WIN32 */
	int locked; /* sanity check */

	void *obj1; /* custom storage area for proxy */
	char name[64];

	int mapcheck_interval;
	time_t next_mapcheck_time; /* when to check again for new mappings */
	struct propertyTable *mappings;

	struct jrun_proxy *pxylist; /* list of sub-proxies */
	struct jrun_proxy *next; /* next proxy in this list */

	/* filter the request by asking proxy if it wants it */
	struct jrun_proxy *(*filterRequest)(struct jrun_proxy *, jrun_request *);

	/* handle a request */
	int (*handleRequest)(struct jrun_proxy *, jrun_request *, const char *target);

	/* free all resources & sub-resources */
	void (*destroy)(struct jrun_proxy *, jrun_request *);

	/* used to add additional servers not defind in jvms.properties,
	   also used to support old-school proxyhost/proxyport/rules.properties */
	int (*addServer)(struct jrun_proxy *, jrun_request *, const char *server);

} jrun_proxy;

void initProxy(jrun_proxy *pxy, jrun_request *r, const char *jrunroot, const char *jvms);

void init_jrun_request(jrun_request *);

#define JRUN_APP_NAME	"JRun 3.0"
#define JRUN_VERSION	"3.00.3664"

/* default property values */
#define JRUN_PROXY_HOST "127.0.0.1"
#define JRUN_PROXY_PORT 8081

/* additional errno's only used internally.  System generated errnos are
 * positive, so there should be no collisions.  If you add more of these,
 * don't forget to add diagnostic messages to the describeerror() function
 * in jrun_property.c 
 */
#define EJRUN_BUSY -10
#define EJRUN_PROTOCOL_ERROR -11

/* other constants */
#define POST_CHUNK_SIZE 40960

/* JRun Proxy Protocol Constants */
enum INETPROXY_CMD {
 PROXY_UNKNOWN     = 100000,
 PROXY_SERVICE			=	0 ,
 PROXY_MAP			=	1,
 PROXY_MAP_CHECK	=	2,
 PROXY_GET_HEADER	=	3,
 PROXY_GET_REALPATH	=	4,
 PROXY_GET_POSTDATA	=	5,    /* read content-length bytes */
 PROXY_PUT_STATUS	=	6,
 PROXY_PUT_HEADER	=	7,
 PROXY_PUT_CLIENT	=	8,
 PROXY_FINISH		=	9,
 PROXY_DESTROYALL	=	10,
 PROXY_DESTROYALL_QUIT=	11,
 PROXY_PING			=	12,
 PROXY_GET_PROPS = 13,
 PROXY_PUT_PROPS = 14,
 PROXY_GET_DATA = 15,   /* read a given number of bytes */
 PROXY_PUT_DATA = 16,   /* return a given number of bytes */
 PROXY_GET_ATTRIBUTE = 17,
 PROXY_SET_ATTRIBUTE = 18,
 PROXY_SET_QUERY_STRING = 19,
 PROXY_SET_REMOTE_HOST = 20,
 PROXY_SET_REMOTE_ADDRESS = 21,
 PROXY_SET_REQUEST_METHOD = 22,
 PROXY_SET_REQUEST_URI = 23,
 PROXY_SET_SERVER_NAME = 24,
 PROXY_SET_SERVER_PORT = 25,
 PROXY_SET_SERVER_PROTOCOL = 26,
 PROXY_SET_PATH_TRANSLATED = 27,
 PROXY_SET_PATH_INFO = 28,
 PROXY_SET_SCRIPT_NAME = 29,
 PROXY_SET_JRUN_TARGET = 30,
 PROXY_SERVICE_30 = 31,
 PROXY_SET_HEADER = 32,
 PROXY_SET_SECURE = 33,
 PROXY_SET_REMOTE_USER = 34,
 PROXY_SET_AUTH_TYPE = 35,
 PROXY_PUT_MORE_DATA = 36,
 PROXY_PUT_HEADER_VALUE = 37,
 PROXY_BUSY = 38
};

typedef struct poolEntry {
	struct poolEntry *next;
	int sock;		/* platform-dependent file descriptor */
} poolEntry;

typedef struct sockPool {
	struct sockaddr_in addr; /* IP and port */
	poolEntry *pool;
} sockPool;

#if defined __cplusplus
}
#endif

#endif
