/* platform.c  -- platform functions common to several connectors
Copyright (c) 1997-2000 Allaire.  All Rights Reserved.

DO NOT REDISTRIBUTE THIS SOFTWARE IN ANY WAY WITHOUT THE EXPRESSED
WRITTEN PERMISSION OF ALLAIRE.
*/

#include "jrun_property.h"

static void jrTraceNoop(jrun_request *r, const char *fmt, ...)
{
}

static int jrConnect(jrun_request *r, struct sockaddr_in *inetaddr)
{
	int i, flag = 1;
	SOCKET sock;

	if ((sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
		r->error = WSAGetLastError();
		r->log(r, "socket() %s", describeError(r->error));
		return -1;
	}

	/* Set some socket options */
    if (setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, (char *) &flag, sizeof(flag)) == -1) {
		r->error = WSAGetLastError();
		r->log(r, "Warning: could not set TCP_NODELAY: %s",
			describeError(r->error));
	}

	do {
		i = connect(sock, (struct sockaddr *) inetaddr, sizeof(struct sockaddr_in) );
	} while (i == -1 && WSAGetLastError() == EINTR);
	if(i == -1)	{
		r->error = WSAGetLastError();
		r->close(r, sock);
		return -1;
	}

	return (int)sock;
}

static int jrRecv(jrun_request *r, int fd, char *buf, int len)
{
	int result;
	do result = recv((SOCKET) fd, buf, len, 0);
	while (result == -1 && WSAGetLastError() == EINTR);
	if (result == -1)
		r->error = WSAGetLastError();
	return result;
}

static int jrSend(jrun_request *r, int fd, const char *buf, int len)
{
	int result;
	do result = send((SOCKET) fd, buf, len, 0);
	while (result == -1 && WSAGetLastError() == EINTR);
	if (result == -1)
		r->error = WSAGetLastError();
	return result;
}

static void jrCloseSocket(jrun_request *r, int fd)
{
#ifdef WIN32
	closesocket((SOCKET) fd);
#else
	close(fd);
#endif
}

static void *jrMalloc(jrun_request *jr, int len)
{
	void *result = malloc(len <= 0 ? 1 : len);
	if (result == null)
		jr->log(jr, "malloc(%d) failed!", len);
	return result;
}

/* ED 5/13/99 use jrMalloc to implement strdup.  Redhat 6.0 strdup
   is a macro which causes problems for us.  This way other platforms
   only need to implement malloc/free. */
static char *jrStrdup(jrun_request *jr, const char *s)
{
	char *result = null;
	if (s != null) {
		int len = strlen(s);
		result = jr->malloc(jr, len+1);
		if (result == null)
			jr->log(jr, "strdup(%d) failed!", len);
		else
			strcpy(result, s);
	}
	return result;
}

static void jrFree(jrun_request *jr, void *p)
{
	if (p == null)
		jr->log(jr, "free(null) !");
	else
		free(p);
}

static void getAttribute(jrun_request *jr, const char *name, char *buf)
{
}

static void setAttribute(jrun_request *jr, const char *name, const char *buf, int len)
{
}

static int getAttributeSize(jrun_request *jr, const char *name)
{
	return 0;
}

static char **getAttributeNames(jrun_request *jr)
{
	return null;
}

void init_jrun_request(jrun_request *r)
{
	memset(r, 0, sizeof(jrun_request));

	r->log = jrTraceNoop;
	r->trace = jrTraceNoop;
	r->malloc = jrMalloc;
	r->free = jrFree;
	r->jstrdup = jrStrdup;
	r->perm_malloc = jrMalloc;
	r->perm_free = jrFree;

	r->connect = jrConnect;
	r->close = jrCloseSocket;
	r->send = jrSend;
	r->recv = jrRecv;

	r->getAttribute = getAttribute;
	r->setAttribute = setAttribute;
	r->getAttributeSize = getAttributeSize;
	r->getAttributeNames = getAttributeNames;
}
