/*
Copyright (c) 1997-2000 Allaire.  All Rights Reserved.

DO NOT REDISTRIBUTE THIS SOFTWARE IN ANY WAY WITHOUT THE EXPRESSED
WRITTEN PERMISSION OF ALLAIRE.
*/

#include <time.h>

#include "extension.h"
#include "jrun_property.h"
#include "jrun_proxy.h"

// Critical section for the file logging
static CRITICAL_SECTION gCS;

// Handle to this DLL
static HANDLE thisHandle;

// Global Variables
static char *		JRunLogPath;
jrun_proxy jrunProxy;
int Verbose = FALSE;
char *errorUrl;

jrun_request global_jr; // use this when you're not inside a real request

/**
 * DllMain()
 *
 * Called when the DLL is loaded and unloaded
 *
 */
BOOL WINAPI 
DllMain(HINSTANCE hInstDLL, DWORD fdwReason, LPVOID lpReserved)
{
	jrun_request *r = &global_jr;
	extern const char *scriptPath;

	init_isapi_request(r, NULL, NULL);

	switch(fdwReason)
	{
		case DLL_PROCESS_ATTACH:
			{
				char *cp, filename[1024];
				const char *serverHost, *s;
				char server[400];
				propTable *	jrunIniProperties;

				//
				// Init the critical section semaphor for logging
				//
				InitializeCriticalSection(&gCS);

				//
				// Get location of .log file
				//
				filename[0] = '\0';
				GetModuleFileName((struct HINSTANCE__ *) hInstDLL, filename, sizeof(filename));
				cp = strrchr(filename, '\\');
				if (cp != NULL) {
					cp[1] = '\0';
					JRunLogPath = (char *) malloc(strlen(filename) + 20);
					strcpy(JRunLogPath, filename);
				} else {
					JRunLogPath = NULL;
					r->log(r, "couldn't find log directory. Logging disabled.");
					filename[0] = '\0';
				}

				// Write JRun version info to log
				r->log(r, "%s ISAPI Extension %s - %s %s", JRUN_APP_NAME,
					JRUN_VERSION, __DATE__, __TIME__);

				//
				// Get location of .ini file
				//
				strcat(filename, "jrun.ini");
	
				//
				// Load ini file
				//
				if((jrunIniProperties = loadProperties(r, filename)) == NULL)
				{
					r->log(r,"Did not find required 'jrun.ini' file in: %s", filename);
					return FALSE;
				}

				//
				// set Verbose mode
				//
				if((s = getPropertyValue(jrunIniProperties, "verbose")) != NULL)	{
					if(strcasecmp(s,"True") == 0) {
						Verbose = TRUE;
					} else {
						Verbose = FALSE;
					}
				}
				r->trace(r, "   verbose is %d", Verbose);
				init_isapi_request(r, NULL, NULL);

				//
				// Init the socket pool
				//
				initProxy(&jrunProxy, r,
					getPropertyValue(jrunIniProperties, "jrun.rootdir"),
					getPropertyValue(jrunIniProperties, "jvmlist"));

				//
				// Get host & port if using old config file
				//
				if((serverHost = getPropertyValue(jrunIniProperties, "proxyhost")) != NULL) {
					strcpy(server, serverHost);
					if((s = getPropertyValue(jrunIniProperties, "proxyport")) == NULL) {
						r->log(r, "Did not find 'port' entry, using '%d'", JRUN_PROXY_PORT);
						sprintf(server, "%.20s:%d", serverHost, JRUN_PROXY_PORT);
					} else {
						sprintf(server, "%.20s:%d", serverHost, atoi(s));
					}
					r->trace(r, "Using proxy %s", server);
					s = getPropertyValue(jrunIniProperties, "rulespath");
					if (s != null) {
						strcat(server, ";");
						strcat(server, s);
						jrunProxy.addServer(&jrunProxy, r, server);
					}
				}

				//
				// Get the errorUrl
				//
				if((errorUrl = (char *) getPropertyValue(jrunIniProperties, "errorurl")) != NULL) {
					r->trace(r, "Using error Url of %s",
						 ( errorUrl == NULL ? "<default>" : errorUrl ) );
					errorUrl = r->jstrdup(r, errorUrl);
				}
							
				// Get the dll path from property file if there
				if((scriptPath = getPropertyValue(jrunIniProperties, "scriptpath")) == NULL) {
					r->log(r, "jrISAPI[filter]: scriptpath property not found, using default /scripts/jrun.dll");
					scriptPath = "/scripts/jrun.dll";
				}
				scriptPath = r->jstrdup(r, scriptPath);
				r->trace(r, "jrISAPI[filter]: scriptPath is %.200s", scriptPath);
				freeProperties(r, jrunIniProperties);
			}
			break;

		case DLL_PROCESS_DETACH:

			r->log(r, "JRun DLL Detaching (normal shutdown)");
			jrunProxy.destroy(&jrunProxy, r);
			if (JRunLogPath != null) {
				free(JRunLogPath);
			}
			break;

		default:
			break;
	}

	return TRUE;
}


/**
 * GetExtensionVersion()
 *
 * Called when the DLL is loaded
 *
 */
BOOL WINAPI 
GetExtensionVersion( HSE_VERSION_INFO *pVer ) 
{
	jrun_request *r = &global_jr;

	// Load description string
	pVer->dwExtensionVersion = MAKELONG( HSE_VERSION_MINOR, HSE_VERSION_MAJOR );
	lstrcpyn( pVer->lpszExtensionDesc, "JRun", HSE_MAX_EXT_DLL_NAME_LEN );

	r->trace(r, "JRun GetExtensionVersion called");

	return TRUE;
}

DWORD WINAPI 
HttpExtensionProc( EXTENSION_CONTROL_BLOCK *pECB )
{
	jrun_request jr, *r = &jr;
	BOOL result;

	__try {

		init_isapi_request(r, pECB, NULL);

		r->trace(r, "In HttpExtensionProc");

		/* handle the comms to the JRun proxy server */
		r->connection = (void *) HSE_STATUS_SUCCESS;

		/* our extension is invoked as "/scripts/jrun.dll/<original-uri> which means
		   the path info is our original URI */
		r->uri = pECB->lpszPathInfo;

		/* build a list of HTTP headers.  The ALL_HTTP variable contains all
		   the http headers in this form:
		      HTTP_HOST:value1
			  HTTP_CONTENT_LENGTH:value2
		   We end up turning this into an array of strings like this
		   
			  { "HOST", "CONTENT-LENGTH", NULL }
		*/

		char *all_http = r->getHeader(r, "ALL_HTTP");
		int hdrCount = 0; char *p;
		for (p=strchr(all_http, '\n'); p != null; p = strchr(p+1,'\n')) {
			hdrCount++;
		}
		r->hdrNames = (char **) r->malloc(r, (1+hdrCount) * sizeof(char *));
		hdrCount = 0;
		for (p = all_http-1; p != null; p = strchr(p, '\n')) {
			if (p > all_http) {
				*p = '\0';
			}
			if (*++p == '\0') {
				break;
			}
			if (!strncmp(p, "HTTP_", 5)) {
				p += 5; // skip "HTTP_"
			}
			r->hdrNames[hdrCount++] = p;
			for (char *u = p; *u != '\0'; u++) {
				if (*u == '_') *u = '-';
				else if (*u == ':') *u-- = '\0';
			}
			p = u+1;
		}
		r->hdrNames[hdrCount] = null;

		if (jrunProxy.handleRequest( &jrunProxy, r, NULL ) != 0) {
			r->trace(r, "handleRequest failed");
			result = HSE_STATUS_ERROR;
		} else {
			result = (BOOL) r->connection;
		}
		r->free(r, r->hdrNames);
		r->free(r, all_http);

	} __except(EXCEPTION_EXECUTE_HANDLER) {

		int code = GetExceptionCode();

		if(code == EXCEPTION_ACCESS_VIOLATION)
			r->log(r, "An ACCESS VIOLATION occured in JRun Connector extension!");
		else
			r->log(r, "An exception occured in JRun Connector extension! (%ld)", code);
		result = HSE_STATUS_ERROR;
	}

	return result;
}

void 
WriteToLog(jrun_request *r, const char *szFormat, ...)
{
	HANDLE hFile;
	char szBuf[4096];
	DWORD dwWritten;
	va_list list;
	char *logname;
	struct tm *tm;
	time_t t;

	t = time(0);
	tm = localtime(&t);
	sprintf(szBuf, "%02d:%02d:%02d %s", tm->tm_hour, tm->tm_min,
		tm->tm_sec, r->stringRep);
	va_start (list, szFormat);
	vsprintf (strchr(szBuf, '\0'), szFormat, list);
	va_end (list);
	strcat(szBuf, "\r\n");

	// Send to dbmon
	OutputDebugString(szBuf);

	// get the current date for the log file name.
	if (JRunLogPath != NULL) {
		logname = strrchr(JRunLogPath,'\\');
		if (logname != NULL) {
			sprintf(logname+1, "jrun%02d%02d%02d.log", tm->tm_year%100,
						tm->tm_mon+1, tm->tm_mday);

			// Send to file
			EnterCriticalSection(&gCS);	

			hFile = CreateFile (JRunLogPath, GENERIC_WRITE, 
								0, NULL, OPEN_ALWAYS, 
								FILE_ATTRIBUTE_NORMAL, NULL);
			if (hFile != INVALID_HANDLE_VALUE) {
				SetFilePointer (hFile, 0, NULL, FILE_END);
				WriteFile (hFile, szBuf, lstrlen (szBuf), &dwWritten, NULL);
				CloseHandle (hFile);
			}

			LeaveCriticalSection(&gCS);	
		}
	}
}
