/*
Copyright (c) 1997-2000 Allaire.  All Rights Reserved.

DO NOT REDISTRIBUTE THIS SOFTWARE IN ANY WAY WITHOUT THE EXPRESSED
WRITTEN PERMISSION OF ALLAIRE.
*/

#include "extension.h"
#include "jrun_property.h"

extern  int Verbose;

static BOOL getHeaderFromAllHttp(jrun_request *jr, 
                                 const char *name, 
                                 const char *all_http,
                                 char *value);
static char *toUpperCase(jrun_request *jr, const char *s);

static char *
getHeader(jrun_request *r, const char *name)
{
	EXTENSION_CONTROL_BLOCK *ECB = (LPEXTENSION_CONTROL_BLOCK) r->request;
	HTTP_FILTER_CONTEXT *pfc = (HTTP_FILTER_CONTEXT *) r->context;
        DWORD len;
        char valbuf[4100]; /* header values can be up to 4,096 bytes long */
	BOOL success;

	/* look for HTTP header values first */
	char *http_name = (char *) r->malloc(r, strlen(name)+6);
	strcpy(http_name, "HTTP_");
	strcat(http_name, name);
	for (char *p = http_name; *p != '\0'; p++) {
		if (*p == '-') {
			*p = '_';
		} else {
			*p = toupper(*p);
		}
	}

	len = sizeof(valbuf);
	if (ECB != NULL) {
		success = ECB->GetServerVariable(ECB->ConnID, http_name, valbuf, &len);
	} else {
		success = pfc->GetServerVariable(pfc, http_name, valbuf, &len);
	}

	if (success) {
		r->trace(r, "   %s: %s", http_name, valbuf);
		r->free(r, http_name);
	    return r->jstrdup(r, valbuf);
	} else {
		r->free(r, http_name);
		http_name = null; /* just in case we try to use it again */
	}

	/* fall through and look for CGI variables */

	/* special hack to change URI to PATH_INFO header name */
	if (strcasecmp(name, "URI") == 0) {
		name = "path_info";
	}

	if (ECB != NULL) {
		/* IIS pulls these values out for us */

		if (strcasecmp(name, "request_method") == 0 && ECB->lpszMethod != NULL) {
			r->trace(r, "   %s: %.200s (special)", name, ECB->lpszMethod);
			return r->jstrdup(r, ECB->lpszMethod);
		}
		if (strcasecmp(name, "query_string") == 0 && ECB->lpszQueryString != NULL) {
			r->trace(r, "   %s: %.200s (special)", name, ECB->lpszQueryString);
			return r->jstrdup(r, ECB->lpszQueryString);
		}
		if (strcasecmp(name, "path_info") == 0 && ECB->lpszPathInfo != NULL) {
			r->trace(r, "   %s: %.200s (special)", name, ECB->lpszPathInfo);
			return r->jstrdup(r, ECB->lpszPathInfo);
		}
		if (strcasecmp(name, "path_translated") == 0 && ECB->lpszPathTranslated != NULL) {
			r->trace(r, "   %s: %.200s (special)", name, ECB->lpszPathTranslated);
			return r->jstrdup(r, ECB->lpszPathTranslated);
		}
		if (strcasecmp(name, "content_type") == 0 && ECB->lpszContentType != NULL) {
			r->trace(r, "   %s: %.200s (special)", name, ECB->lpszContentType);
			return r->jstrdup(r, ECB->lpszContentType);
		}
	}

	len = sizeof(valbuf);
	if (ECB != NULL) {
		success = ECB->GetServerVariable(ECB->ConnID, (char *) name, valbuf, &len);
	} else {
		success = pfc->GetServerVariable(pfc, (char *) name, valbuf, &len);
	}

        /* Work around for headers set by other filters */
        if (! success) {
          /* Get the ALL_HTTP header, which should contain all the headers */
          if (ECB != NULL) {
            success = ECB->GetServerVariable(ECB->ConnID, "ALL_HTTP", valbuf, &len);
          } else {
            success = pfc->GetServerVariable(pfc, "ALL_HTTP", valbuf, &len);
          }

          if (success) {
            success = getHeaderFromAllHttp(r, name, valbuf, valbuf);
          }
        }

	if (success) {
		r->trace(r, "   %s: %s", name, valbuf);
		return r->jstrdup(r, valbuf);
	}
	
	if (GetLastError() == ERROR_MORE_DATA) {
		valbuf[sizeof(valbuf)-1] = '\0';	
		r->log(r, "   %s: %.200s<truncated>", name, valbuf);
		return r->jstrdup(r, valbuf);
	}

	r->trace(r, "   %s: <null>", name);
	return NULL;
}

static BOOL
getHeaderFromAllHttp(jrun_request *r, const char *name, const char *all_http, char *value)
{
  BOOL success = false;
  /* All the headers in ALL_HTTP are uppercase, so convert
     the header to uppercase. */
  char *uppercase_name = toUpperCase(r, name);

  /* Search for the header in the ALL_HTTP value. */
  char *header = strstr(all_http, uppercase_name);
  
  if (header != NULL) {

    char *previousChar = (header - 1);
    char *nextChar = header + strlen((char *) name);

    if (((*previousChar == '\n') || 
         (*previousChar == ' ')) &&
        (*nextChar == ':')) {
      /* Add 1 to account for the colon separator. */
      char *valueBegin = nextChar + 1;
      char *valueEnd = strchr(valueBegin, '\n');

      int length;

      if ( valueEnd == NULL ) {
        length = strlen(valueBegin);
      } else {
        length = valueEnd - valueBegin;
      }
    
      value = (char *) r->malloc(r, length + 1);

      strncpy(value, valueBegin, length);

      success = true;
    }
  }

  return success;
}

/* conversion from xxx to XXX, not in place */
static char *
toUpperCase(jrun_request *r, const char *s)
{
	int i, len = strlen(s);
	char *uppercase = (char *) r->malloc(r, len+1);

	for(i = 0; i < len; i++) {
		if(s[i] == '-')
			uppercase[i] = '_';
		else
			uppercase[i] = toupper(s[i]);
	}
	
	uppercase[i] = '\0';

	return uppercase;
}

static char *
getRealPath(jrun_request *r, const char *path)
{
	EXTENSION_CONTROL_BLOCK *ECB = (LPEXTENSION_CONTROL_BLOCK) r->request;

	DWORD len = 4100;
	
	char *realPath = (char*) r->malloc( r, len );
	if (realPath != NULL) {
		lstrcpy(realPath, path);

		ECB->ServerSupportFunction(ECB->ConnID,
					   HSE_REQ_MAP_URL_TO_PATH,
					   realPath,
					   &len,
					   &len);

		realPath[len] = '\0';
	}
	r->trace(r, "   getRealPath(%s) = %s", path, realPath?realPath:"<null>");
    return realPath;
}


static int
getPostData(jrun_request *r, char *buf, int clen)
{
	EXTENSION_CONTROL_BLOCK *ECB = (LPEXTENSION_CONTROL_BLOCK) r->request;

 	DWORD   cbQuery = 0;
	BOOL	prematureAbort = FALSE;
	int		totalBytesSent = 0;
	DWORD avail = 0, ecb_avail = 0;

	if (r->shouldRead && ECB->cbTotalBytes > 0) {
		if (!r->shouldReadMore) {
			ecb_avail = ECB->cbAvailable - r->offset;  // skip r->offset bytes
			r->trace(r, "   getPostData has %d available", ecb_avail );
			// get first ecb_avail bytes, or clen, whichever is smaller
			avail = ecb_avail < (DWORD)clen ? ecb_avail : (DWORD)clen;
			memcpy(buf, ECB->lpbData + r->offset, avail);
			r->offset += avail;
			if ((DWORD)r->offset >= ECB->cbAvailable) {
				// done reading initial chunk
				if (ECB->cbAvailable == ECB->cbTotalBytes)
					r->shouldRead = 0; // EOF
				else
					r->shouldReadMore = 1; // read more data from sock
			}
		} else {
			// read remaining data from socket
			avail = clen;
			r->trace(r, "   getPostData reading up to %d bytes", avail );
			if (ECB->ReadClient(ECB->ConnID, buf, &avail)) {
				// success
				if (avail == 0) {
					// eof
					r->shouldRead = 0;
				}
			} else {
				// failure
				if (avail == 0) avail = -1;
				r->shouldRead = 0;
			}
		}
	} else {
		// no data to read.
		r->shouldRead = 0;
	}
	return avail;
}
		
/**
  write HTTP header back to browser.  szStatus is the status
  message, such as "200 OK\r\n".  szStatus and szHeaders both
  should include the correct EOL sequences.
*/

static void
setStatus(jrun_request *r, int statusCode,  const char *szStatus)
{
	EXTENSION_CONTROL_BLOCK *ECB = (LPEXTENSION_CONTROL_BLOCK) r->request;

	unsigned long len = 0;

	extern char *errorUrl;

	if (statusCode == 503 && errorUrl != NULL) { // assume proxy error
		DWORD len = strlen( errorUrl );
		r->noaction = 1;  // ignore content to follow

		r->trace(r, "ISAPI redirect to %s", errorUrl);
		if (!ECB->ServerSupportFunction(ECB->ConnID,
			HSE_REQ_SEND_URL_REDIRECT_RESP,	errorUrl, &len, 0))
			r->log(r, "warning: ISAPI send redirect failed: %s", describeError(GetLastError()));
		return;
	}

	if (szStatus == NULL) szStatus = "";
	r->reason = (char *) r->malloc(r, strlen(szStatus) + 20);
	sprintf(r->reason, "%d %s", statusCode, szStatus);

	r->headersSize = 1024;
	r->headers = (char *) r->malloc(r, r->headersSize);
	r->headers[0] = '\0';
	r->headersEnd = r->headers;
}

static void
addHeader(jrun_request *r, const char *name, const char *value)
{
	EXTENSION_CONTROL_BLOCK *ECB = (LPEXTENSION_CONTROL_BLOCK) r->request;
	int space = (r->headersEnd-r->headers) + strlen(name) + 2 + strlen(value) + 1;

	if (space > r->headersSize) {
		/* grow more memory */
		r->headersSize = 2*space;
		char *temp = (char *) r->malloc(r, r->headersSize);
		strcpy(temp, r->headers);
		r->headersEnd = temp + (r->headersEnd - r->headers);
		r->free(r, r->headers);
		r->headers = temp;
	}

	if (strcasecmp(name, "connection") == 0) {
		if (strcasecmp(value, "keep-alive") == 0) {
			r->connection = (void *) HSE_STATUS_SUCCESS_AND_KEEP_CONN;
		} else {
			r->connection = (void *) HSE_STATUS_SUCCESS;
		}
		sprintf(r->headersEnd, "%s: %s\r\n", name, value);
		r->trace(r, "   (noted keep-alive) %s: %s", name, value);

	} else if (strcasecmp(name, "date") == 0) {
		/* ingore date headers, IIS adds its own and we can't change it */
		r->trace(r, "   (suppressed) %s: %s", name, value);

	} else if (strcasecmp(name, "server") == 0) {
		/* ignore this header -- IIS sets it automatically */
		r->trace(r, "   (suppressed) %s: %s", name, value);

	} else {
		sprintf(r->headersEnd, "%s: %s\r\n", name, value);
		r->trace(r, "   (noted) %s: %s", name, value);
	}
	r->headersEnd = strchr(r->headersEnd, '\0');
}


/* 
 * writeClient
 *
 * return 0 == success, -1 == failure
 *
 */
static int
writeClient(jrun_request *r, const char *data, int len)
{
	EXTENSION_CONTROL_BLOCK *ECB = (LPEXTENSION_CONTROL_BLOCK) r->request;
	int writeLen = len;

	if (!r->noaction && !r->committed) {
		strcpy(r->headersEnd, "\r\n");
		DWORD hlen = r->headersEnd - r->headers;
		BOOL result = ECB->ServerSupportFunction(ECB->ConnID,
			       HSE_REQ_SEND_RESPONSE_HEADER,
			       r->reason, &hlen, (LPDWORD) r->headers);

		if (result != TRUE) {
			r->trace(r, "     HSE_REQ_SEND_RESPONSE_HEADER failed: %d", GetLastError());
			r->connection = (void *) HSE_STATUS_ERROR;
			r->noaction = 1;
		}
	}

	if (!r->noaction) {
		r->trace(r, "   writeClient: sending %ld bytes", len );
		ECB->WriteClient(ECB->ConnID, (char *) data, (unsigned long *) &writeLen, HSE_IO_SYNC);
		if(len != writeLen) {
			r->log(r, "   writeClient failed: %d", GetLastError());
			return -1;
		}
	}
	return 0;
}

void
init_isapi_request(jrun_request *r, EXTENSION_CONTROL_BLOCK *pECB,
				   HTTP_FILTER_CONTEXT *pFC)
{
	init_jrun_request(r);
	r->request = (void *) pECB;
	r->context = (void *) pFC; 

	if (pECB != NULL) {
		sprintf(r->stringRep, "jrISAPI[%d] ", ((int)pECB->ConnID) % 10000);
	} else if (pFC != NULL) {
		sprintf(r->stringRep, "jrISAPI[filter] ");
	} else {
		strcpy(r->stringRep, "jrISAPI[init] ");
	}

	r->shouldRead = 1;
	r->shouldReadMore = 0;
	r->offset = 0;

	r->getHeader = getHeader;
	r->getRealPath = getRealPath;
	r->getPostData = getPostData;
	r->setStatus = setStatus;
	r->addHeader = addHeader;
	r->writeClient = writeClient;

	r->log = WriteToLog;
	if (Verbose) {
		r->trace = WriteToLog;
	}
}
