/*
Copyright (c) 1997-2000 Allaire.  All Rights Reserved.

DO NOT REDISTRIBUTE THIS SOFTWARE IN ANY WAY WITHOUT THE EXPRESSED
WRITTEN PERMISSION OF ALLAIRE.
*/

#include "extension.h"
#include "jrun_property.h"

int	jrNSAPIVerbose = 0;
static struct jrun_proxy jrunproxy;
static jrun_request globalRequest;

static int loglevel = LOG_INFORM;

char *errorUrl;

NSAPI_PUBLIC void
jrunend(void *param)
{
	jrun_request *r = &globalRequest;

	loglevel = LOG_INFORM;
	r->log(r, "JRun connector exiting");
	jrunproxy.destroy(&jrunproxy, r);
}

/* Plug-in init code */
NSAPI_PUBLIC int
jruninit(pblock *pb, Session *sn, Request *rq)
{
	char *s, *p;
	jrun_request *r = &globalRequest;
	char server[512];

		/* Check verbose mode */
	s = pblock_findval("verbose", pb);
	if(s != NULL && strcmp(s, "true") == 0)
		jrNSAPIVerbose = TRUE;

	/* initialize the global request so we can call NSAPI functions
		outside the context of a request */
	init_nsapi_request(r, NULL, NULL, NULL);

	/* Print version info */

	loglevel = LOG_INFORM;
	r->log(r, "%s %s NSAPI Extension - %s %s", JRUN_APP_NAME, JRUN_VERSION,
		__DATE__, __TIME__);

	loglevel = LOG_WARN;

	/* init the JRun connection */
	initProxy(&jrunproxy, r, pblock_findval("jrun.rootdir", pb),
		pblock_findval("jvmlist", pb));

	/* Get the jrun proxy info.  use jrunproxy first, or proxyhost/proxyport */
	if((s = pblock_findval("proxyhost", pb)) != NULL) {
		p = pblock_findval("proxyport", pb);
		sprintf(server, "%.15s:%d", s, p != NULL ? atoi(p) : JRUN_PROXY_PORT);
	}
	r->trace(r, "Server: %s", server);
	p = pblock_findval("rulespath", pb);
	if (p != null) {
		strcat(server, ";");
		strcat(server, p);
		jrunproxy.addServer(&jrunproxy, r, server);
	}

	/* get the error url property, if it exists */
	if((s = pblock_findval("errorurl", pb)) != NULL) {
		errorUrl = s;
	}

	daemon_atrestart(jrunend, NULL);

	return REQ_PROCEED;
}


/* Main service routine */
NSAPI_PUBLIC   int 
jrunservice(pblock *pb, Session *sn, Request *rq)
{
		jrun_request r;
		int error = REQ_PROCEED;
		const char *ctype;
		struct pb_entry *p;
		int i, j;

		ctype = pblock_findval("content-type", rq->srvhdrs);
		if (ctype != NULL && strncmp(ctype, "jrun-internal/", 14) == 0) {
			/* this request is for us, but lets remove the internal content-type */
			param_free(pblock_remove("content-type", rq->srvhdrs));
		} else {
			/* this request is not for us */
			return REQ_NOACTION;
		}

		init_nsapi_request(&r, pb, sn, rq);
#ifdef XP_WIN32
	__try {
#endif
		r.shouldReadMore = 1;

		j = 0;
		/* got header traversal code from:
		http://developer.netscape.com/support/faqs/index.html?content=champions/nsapi.html */
		for (i=0; i<rq->headers->hsize; i++) {
			for (p = rq->headers->ht[i]; p != null; p = p->next) {
				j++;
			}
		}
		r.hdrNames = r.malloc(&r, (1+j) * sizeof (char *));
		j = 0;
		for (i=0; i<rq->headers->hsize; i++) {
			for (p = rq->headers->ht[i]; p != null; p = p->next) {
				r.hdrNames[j++] = p->param->name;
			}
		}
		r.hdrNames[j] = NULL;

		/* handle the comms to the JRun proxy server */
		if (jrunproxy.handleRequest(&jrunproxy, &r, NULL) != 0) {
			r.trace(&r, "handleRequest failed");
			error = REQ_ABORTED;
		}
		r.free(&r, r.hdrNames);

#ifdef XP_WIN32
	} __except(EXCEPTION_EXECUTE_HANDLER) {

		int code = GetExceptionCode();

		if(code == EXCEPTION_ACCESS_VIOLATION)  {
			r.log(&r, "An ACCESS VIOLATION occured in JRun Connector extension!" );
		} else {
			r.log(&r, "An exception occured in JRun Connector extension! (%ld)", code);
		}
		error = REQ_ABORTED;

		if (r.hdrNames != NULL) {
			r.free(&r, r.hdrNames);
		}
	}
#endif

	return error;
}

/**
 * Fix the URI.  Here we figure out what the URL would look
 * like that gave us the matched document.  That way JRun 2.3 can
 * do its rule mapping on the URL.  For example,
 *    uri    = /
 *    base   = d:/nes/suitespot/docs
 *    path   = d:/nes/suitespot/docs/index.html
 *    newuri = /index.html
 *
 * The for loop lets us do this even when base only accounts for
 * part of the URI (you can't simply append the path remainder
 * to the URI).  For example:
 *    uri    = /virtual/home
 *    base   = d:/virtual
 *    path   = d:/virtual/home/index.html
 *    newuri = /virtual/home/index.html
 */
static char *fix_uri(jrun_request *r,
					 const char *uri, const char *path, const char *base)
{
	char *newuri;
	const char *doc;

	r->trace(r, "   base  %s", base ? base : "null");
	r->trace(r, "   uri   %s", uri  ? uri  : "null");
	r->trace(r, "   path  %s", path ? path : "null");

	if (base != NULL && strncmp(path, base, strlen(base)) == 0) {
		const char *u;
		/* doc = whats after ntrans-base */
		doc = path + strlen(base);
		/* remove whatever overlaps with the URI */
		for (u = uri; u != NULL; u = strchr(u+1, '/')) {
			if (strncmp(u, doc, strlen(u)) == 0) {
				doc += strlen(u);
				break;
			}
		}
	} else {
		/* path is wacked, just take document name */
		doc = strrchr(path, '/');
		if (doc == NULL) doc = "";
	}
	if (doc[0] == '/') doc++;
	newuri = r->malloc(r, strlen(uri) + strlen(doc) + 1);
	strcpy(newuri, uri);
	strcat(newuri, doc);
	return newuri;
}

NSAPI_PUBLIC int
jrunfilter(pblock *pb, Session *sn, Request *rq)
{
		char * newMapping = NULL;
		char *path, *ctype, *base;
		jrun_request r;
		jrun_proxy *target;

		ctype = pblock_findval("content-type", rq->srvhdrs);
		if (ctype != NULL && strncmp(ctype, "jrun-internal/", 14) == 0) {
			/* we have already filtered this request. */
			return REQ_NOACTION;
		}

		/* init a request object for logging and memory allocation */
		init_nsapi_request(&r, pb, sn, rq);

		path = pblock_findval("ppath", rq->vars);
		if (path != NULL) {
			/* NameTrans directive.  Add jrun object to active objects,
			   and do filtering now, in case they don't have a PathCheck directive. */
			pblock_nvinsert("name", "jrun", rq->vars);
		} else {
			/* PathCheck directive.  do filtering now. */
			path = pblock_findval("path", rq->vars);
		}

		r.path = path;
		if (r.uri == NULL) {
			r.trace(&r, "no URI, declining request");
			return REQ_NOACTION;
		}

		base = pblock_findval("ntrans-base", rq->vars);
#ifdef XP_WIN32
	__try {
#endif

		target = jrunproxy.filterRequest(&jrunproxy, &r);
		if (target != NULL) {
			if (r.uri[strlen(r.uri)-1] == '/') {
				/* uri was a directory, but we matched a file.  so add that
					filename to the URI so JRun knows what to do with it */
				char *newuri = fix_uri(&r, r.uri, r.path, base);
				pblock_remove("uri", rq->reqpb);
				pblock_nvinsert("uri", newuri, rq->reqpb);
			}
			pblock_nvinsert("content-type", "jrun-internal/target", rq->srvhdrs);
			/* allow other NameTrans and PathCheck directives to run */
			return REQ_NOACTION;
		}

#ifdef XP_WIN32
	} __except(EXCEPTION_EXECUTE_HANDLER) {

		int code = GetExceptionCode();

		if(code == EXCEPTION_ACCESS_VIOLATION)
			r.log(&r,"An ACCESS VIOLATION occured in JRun Connector filter!" );
		else
			r.log(&r,"An exception occured in JRun Connector filter! (%ld)", code);
	}
#endif

    return REQ_NOACTION;
}

void
WriteToLog(jrun_request *r, const char *szFormat, ...)
{
	va_list list;
	char szBuf[1024];
	Session *sn = (Session *) r->connection;
	Request *rq = (Request *) r->request;

	va_start (list, szFormat);
	vsprintf (szBuf, szFormat, list);
	va_end (list);

	/* this goes into NES's log */
	log_error(loglevel, r->stringRep, sn, rq, "%s", szBuf);

#ifdef XP_WIN32
	/* on NT, send string to the debug monitor */
	strcat(szBuf, "\r\n");
	OutputDebugString(szBuf);
#endif
}
