/*
Copyright (c) 1997-2000 Allaire.  All Rights Reserved.

DO NOT REDISTRIBUTE THIS SOFTWARE IN ANY WAY WITHOUT THE EXPRESSED
WRITTEN PERMISSION OF ALLAIRE.
*/

#include "extension.h"
#include "interface.h"
#include "jrun_property.h"

static char *getOtherHeader(jrun_request *r, const char *headerName);
static char *toUpperCaseWithUnderscores(jrun_request *r, const char *s);
static char *toUpperCaseWithDashes(jrun_request *r, const char *s);
static char *toLowerCaseWithUnderscores(jrun_request *r, const char *s);
static char *toLowerCaseWithDashes(jrun_request *r, const char *s);

static void *
doMalloc(jrun_request *jr, int len)
{
	void *result = MALLOC(len <= 0 ? 1 : len);
	if (result == NULL)
		jr->log(jr, "malloc(%d) failed!", len);
	return result;
}

static void *
doPermMalloc(jrun_request *jr, int len)
{
	void *result = PERM_MALLOC(len <= 0 ? 1 : len);
	if (result == NULL)
		jr->log(jr, "permMalloc(%d) failed!", len);
	return result;
}

static void
doFree(jrun_request *jr, void *p)
{
	if (p == NULL)
		jr->log(jr, "FREE(NULL)!");
	else
		FREE(p);
}

static void
doPermFree(jrun_request *jr, void *p)
{
	if (p == NULL)
		jr->log(jr, "FREE(NULL)!");
	else
		PERM_FREE(p);
}

static char *
getRealPath(jrun_request *r, const char *pathName)
{
	Session *sn = (Session *) r->connection;
	Request *rq = (Request *) r->request;
	const char *method;

	const char *uri = pblock_findval("uri", rq->reqpb);
	char *translated;

	if (uri != NULL && strcmp(pathName, uri) == 0) {
		translated = pblock_findval("path", rq->vars);
		method = "shortcut";
	} else {
		translated = servact_translate_uri((char *) pathName, sn);
		method = "translated";
	}

	/* known problem: 
	   servact_translate_uri() only gives the right answer when
	   virtual hosts are not in effect.  Virtual host nametrans
	   directives are not applied */
	
	if (translated != NULL) {
		r->trace(r, "   %s getRealPath(%.200s) = %.200s", method,
			pathName, translated);
		return r->jstrdup(r, translated);
	} else {
		r->trace(r, "   %s getRealPath(%.200s) = null", method,	pathName);
		return NULL;
	}
}

static char *
getHeader(jrun_request *r, const char *headerName)
{
	Session *sn = (Session *) r->connection;
	Request *rq = (Request *) r->request;

	int i = 0;
	int newHeaderNameConv = FALSE;
	char *newHeaderName = NULL;
	char *requestedHeader = NULL;

	/* shortcut for http headers */
	if (rq != null && rq->headers != null) {
		requestedHeader = pblock_findval((char *) headerName, rq->headers);
		if (requestedHeader != null) {
			r->trace(r, "   %.100s: %.200s", headerName, requestedHeader);
			return r->jstrdup(r, requestedHeader);
		}
	}

	/* SERVER_PORT_SECURE */
	if(strcasecmp((char *) headerName, "SERVER_PORT_SECURE") == 0) {
		char *result = (char *) r->malloc(r, 2);
		sprintf(result, "%d", security_active ? 1 : 0);
		r->trace(r, "   SERVER_SUPPORT_SECURE: %d (custom)", security_active);
		return result;
	}

	/* PATH_TRANSLATED */
	if(strcasecmp((char *) headerName, "PATH_TRANSLATED") == 0) {
		char *pathinfo = pblock_findval("path-info", rq->vars);
		char *realpath;
		if(NULL != pathinfo) {
			realpath = servact_translate_uri(pathinfo, sn);
			if (realpath != NULL)
				return r->jstrdup(r, realpath);
			else
				return NULL;
		} else
			return NULL;
	}

	/* REMOTE_HOST */
	if(strcasecmp((char *) headerName, "REMOTE_HOST") == 0) {
		char * host = session_dns(sn);
		if(host != NULL) {
			r->trace(r, "   %s: %.100s", headerName, host);
			return r->jstrdup(r, host);
		} else
			headerName = "REMOTE_ADDR";
	}

	/* SERVER_NAME */
	if(strcasecmp((char *) headerName, "SERVER_NAME") == 0) {
		r->trace(r, "   %s: %.100s", headerName, server_hostname);
		return r->jstrdup(r, server_hostname);
	}

	/* SERVER_PORT */
	if(strcasecmp((char *) headerName, "SERVER_PORT") == 0) {
		char *buf = r->malloc(r, 16);
		sprintf(buf, "%d", conf_getglobals()->Vport);
		r->trace(r, "   %s: %s", headerName, buf);
		return buf;
	}

	/* ALL OTHERS */

        /* Lookup the header as is first. */
        requestedHeader = getOtherHeader(r, headerName);

        /* Next try finding the Netscape equivalent or Netscape
           mangled header */
        if (requestedHeader == NULL || requestedHeader[0] == '\0') {
          for(i=0;i<NESEQUIV_HEADER_COUNT;i++) {
		if(strcasecmp((char *) headerName, nesequivHeaders[i].name) == 0) {
			newHeaderName = nesequivHeaders[i].equiv;
			break;
		}
          }

          if(newHeaderName == NULL) {
            newHeaderName = toUpperCaseWithUnderscores(r, headerName);
            newHeaderNameConv = TRUE;
          }

          requestedHeader = getOtherHeader(r, newHeaderName);

          if(requestedHeader == NULL || requestedHeader[0] == '\0') {
            newHeaderName = toUpperCaseWithDashes(r, headerName);
            newHeaderNameConv = TRUE;
          }

          requestedHeader = getOtherHeader(r, newHeaderName);

          if(requestedHeader == NULL || requestedHeader[0] == '\0') {
            newHeaderName = toLowerCaseWithUnderscores(r, headerName);
            newHeaderNameConv = TRUE;
          }

          requestedHeader = getOtherHeader(r, newHeaderName);

          if(requestedHeader == NULL || requestedHeader[0] == '\0') {
            newHeaderName = toLowerCaseWithDashes(r, headerName);
            newHeaderNameConv = TRUE;
          }

          requestedHeader = getOtherHeader(r, newHeaderName);

          if (newHeaderNameConv == TRUE)
            r->free(r, newHeaderName);
        }

	return r->jstrdup(r, requestedHeader);
}

static char *
getOtherHeader(jrun_request *r, const char *headerName)
{
	Session *sn = (Session *) r->connection;
	Request *rq = (Request *) r->request;

	char *requestedHeader = NULL;
	char *foundin;

	/* The "method", "uri" and "protocol" of request */
	foundin = "rq->reqpb";
	requestedHeader = pblock_findval((char *) headerName, rq->reqpb);

	if ( requestedHeader == NULL || requestedHeader[0] == '\0' ) {

		/* Server working variables */
		foundin = "rq->vars";
		requestedHeader = pblock_findval((char *) headerName, rq->vars);

		if ( requestedHeader == NULL || requestedHeader[0] == '\0' ) {

			/* client information */
			foundin = "sn->client";
			requestedHeader = pblock_findval((char *) headerName, sn->client);

			if ( requestedHeader == NULL || requestedHeader[0] == '\0' ) {

				foundin = "not";
				requestedHeader = "";
			}
		}
	}

        if ( requestedHeader[0] != '\0' )
        {
          r->trace(r, "   %.100s: %.200s [Found in: %s]", headerName, requestedHeader, foundin);
        }

        return r->jstrdup(r, requestedHeader);
}

static int
getPostData(jrun_request *r, char *buf, int clen)
{
	Session *sn = (Session *) r->connection;
	Request *rq = (Request *) r->request;
	int avail = 0;

	/* see this URL for when to insert data-removed to rq->vars
http://developer.netscape.com/support/faqs/index.html?content=champions/nsapi.html */

#if defined(XP_NETWARE) || defined(XP_WIN32)
	/* ED 2/15/98 - Novell server doesn't like to handle large chunks of data */
	if (clen > 8192) clen = 8192;
#endif

	/* first we'll check to see if any has already been read */
	if (clen > 0 && r->shouldReadMore) {
		avail = sn->inbuf->cursize - sn->inbuf->pos;
		if (avail == 0) {
			avail = netbuf_grab(sn->inbuf, clen);
			if (avail == IO_ERROR || avail == IO_EOF) {
				r->log(r, "getPostData: Premature read abort on post data");
				r->shouldReadMore = 0;
				return -1;
			}
			r->trace(r, "   Read %d bytes of post data", avail);
		} else {
			r->trace(r, "   inbuf has %d bytes ready", avail);
		}
		if (avail > 0) {
			if (avail > clen) avail = clen;
			memcpy(buf, sn->inbuf->inbuf + sn->inbuf->pos, avail);
			sn->inbuf->pos += avail;
			clen -= avail;
		}
	}
	return avail;
}

static void
setStatus(jrun_request *jr, int statusCode,  const char *szStatus)
{
	Session *sn = (Session *) jr->connection;
	Request *rq = (Request *) jr->request;
	extern char *errorUrl;

	if (statusCode == 503 && errorUrl != NULL) {
		/* do a redirect */
		statusCode = 302;
		pblock_nvinsert("location", errorUrl, rq->srvhdrs);
		protocol_status(sn, rq, PROTOCOL_REDIRECT, (char *) szStatus);
	} else {
					/* set the status code and reason string */
		protocol_status(sn, rq, statusCode, (char *) szStatus);
	}
}

static void
addHeader(jrun_request *jr, const char *constname, const char *constvalue)
{
	Request *rq = (Request *) jr->request;

	char *name = (char *) constname;
	char *value = (char *) constvalue;

			if (strcasecmp(name, "set-cookie") == 0) {
				jr->trace(jr, "   (insert)  %s: %.200s", name, value);
				pblock_nvinsert("set-cookie", value, rq->srvhdrs);

			} else if (strcasecmp(name, "server") == 0) {
				/* ignore this header -- Netscape sets it automatically */
				jr->trace(jr, "   (suppressed) %s: %.80s", name, value);

			} else if (strcasecmp(name, "date") == 0) {
				/* ignore the Date: header if the client sets it */
				jr->trace(jr, "   (suppressed) %s: %.20s", name, value);

				/* catch & replace the following headers, because they do 
				special things.  NES needs them to be in lower case. */
			} else if (strcasecmp(name, "connection") == 0) {
				jr->trace(jr, "   (special) %s: %.40s", name, value);
				pblock_remove("connection", rq->srvhdrs);
				pblock_nvinsert("connection", value, rq->srvhdrs);

			} else if (strcasecmp(name, "content-length") == 0) {
				jr->trace(jr, "   (special) %s: %.20s", name, value);
				pblock_remove("content-length", rq->srvhdrs);
				pblock_nninsert("content-length", atoi(value), rq->srvhdrs);

			} else if (strcasecmp(name, "content-type") == 0) {
				char *s;
				jr->trace(jr, "   (special) content-type: %.100s", value);
				/* ED 5/3/99 only lower-case up to ';' to handle this header:
				       "content-type: multipart/x-mixed-replace; <boundary string>" */
				value = jr->jstrdup(jr, constvalue); /* don't modify const memory */
				for (s=value; *s != '\0' && *s != ';'; s++)
					*s = tolower(*s);
				pblock_remove("content-type", rq->srvhdrs);
				pblock_nvinsert("content-type", value, rq->srvhdrs);

			} else {
				/* put it in r->headers_out */
				pblock_remove(name, rq->srvhdrs);
				pblock_nvinsert(name, value, rq->srvhdrs);
				jr->trace(jr, "   (replace) %.80s: %.200s", name, value);
			}
}

static int
writeClient(jrun_request *r, const char *data, int len)
{
	Session *sn = (Session *) r->connection;
	Request *rq = (Request *) r->request;
	
	if (!r->committed && !r->noaction) {
		switch (protocol_start_response(sn, rq)) {
		case REQ_NOACTION:
			/* method was probably a HEAD, so we don't need to
			do anything else after this.  Set a flag in jrun_request */
			r->trace(r, "   start_response = REQ_NOACTION");
			r->noaction = 1;
			break;
		case REQ_ABORTED:
			/* something bad happened and we should stop the request */
			r->trace(r, "   start_response = REQ_ABORTED");
			r->noaction = 1;
			break;
		default: /* REQ_PROCEED */
			/* okay */
			break;
		}
	}

	if(!r->noaction && net_write(sn->csd, (char *) data, (int) len)  == IO_ERROR ) {
		r->error = errno;
		return -1;
	} else {
		return 0;
	}
}

static char *
toUpperCaseWithDashes(jrun_request *jr, const char *s)
{
	int i, len = strlen(s);
	char *r = jr->malloc(jr, len+1);

	for(i=0;i<len;i++) {
		if(s[i] == '_')
			r[i] = '-';
		else
			r[i] = toupper(s[i]);
	}
	
	r[i] = '\0';

	return r;
}

static char *
toUpperCaseWithUnderscores(jrun_request *jr, const char *s)
{
	int i, len = strlen(s);
	char *r = jr->malloc(jr, len+1);

	for(i=0;i<len;i++) {
		if(s[i] == '-')
			r[i] = '_';
		else
			r[i] = toupper(s[i]);
	}
	
	r[i] = '\0';

	return r;
}

static char *
toLowerCaseWithDashes(jrun_request *jr, const char *s)
{
	int i, len = strlen(s);
	char *r = jr->malloc(jr, len+1);

	for(i=0;i<len;i++) {
		if(s[i] == '_')
			r[i] = '-';
		else
			r[i] = tolower(s[i]);
	}
	
	r[i] = '\0';

	return r;
}

static char *
toLowerCaseWithUnderscores(jrun_request *jr, const char *s)
{
	int i, len = strlen(s);
	char *r = jr->malloc(jr, len+1);

	for(i=0;i<len;i++) {
		if(s[i] == '-')
			r[i] = '_';
		else
			r[i] = tolower(s[i]);
	}
	
	r[i] = '\0';

	return r;
}

static int nsapi_connect(jrun_request *r, struct sockaddr_in *inetaddr)
{
	int i;
	SYS_NETFD sock;
	int flag = 1;

	if ((sock = net_socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == SYS_NET_ERRORFD) {
		r->error = errno;
		r->log(r, "socket() %s", describeError(r->error));
		return -1;
	}


	/* Set some socket options */
    if (net_setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, &flag, sizeof(flag)) == -1) {
		r->error = WSAGetLastError();
		r->log(r, "Warning: could not set TCP_NODELAY: %s",
			describeError(r->error));
	}

	do {
		i = net_connect(sock, (struct sockaddr *) inetaddr, sizeof(struct sockaddr_in) );
	} while (i == -1 && errno == EINTR);

	if (i == -1) {
		r->error = errno;
		net_close(sock);
		return -1;
	}
	return (int)sock;
}

static int nsapi_recv(jrun_request *r, int fd, char *buf, int size)
{
	int result;

	do {
		result = net_read((SYS_NETFD) fd, buf, size, NET_INFINITE_TIMEOUT);
	} while (result < 0 && errno == EINTR);
	if (result < 0)
		r->error = errno;
	return result;
}

static int nsapi_send(jrun_request *r, int fd, const char *buf, int size)
{
	int result = net_write((SYS_NETFD) fd, (char *) buf, size);
	if (result < 0)
		r->error = errno;
	return result;
}

static void nsapi_close(jrun_request *r, int fd)
{
	net_close((SYS_NETFD) fd);
}

/* init jrun_request object with function pointers */
void
init_nsapi_request(jrun_request *r, pblock *pb, Session *sn, Request *rq)
{
	extern int jrNSAPIVerbose; /* defined in extension.c */

	init_jrun_request(r);
	r->context = (void *) pb;
	r->connection = (void *) sn;
	r->request = (void *) rq;
	r->noaction = 0;

	if (sn != NULL) {
		struct sockaddr_in addr;
		int i = sizeof(addr);
		net_getpeername(sn->csd, (struct sockaddr *) &addr, &i);
		i = ntohs(addr.sin_port);
		sprintf(r->stringRep, "jrNSAPI[%d] ", i);
	} else {
		strcpy(r->stringRep, "jrNSAPI[init] ");
	}

	r->getHeader = getHeader;
	r->getRealPath = getRealPath;
	r->getPostData = getPostData;
	r->setStatus = setStatus;
	r->addHeader = addHeader;
	r->writeClient = writeClient;

	r->malloc = doMalloc;
	r->free = doFree;
	r->perm_malloc = doPermMalloc;
	r->perm_free = doPermFree;

	r->log = WriteToLog;
	if (jrNSAPIVerbose)
		r->trace = WriteToLog;

	r->connect = nsapi_connect;
	r->close = nsapi_close;
	r->send = nsapi_send;
	r->recv = nsapi_recv;

	if (rq != NULL) {
		r->uri = pblock_findval("uri", rq->reqpb);
	}
}
